__author__ = 'JB'


import itertools

from pyklip.kpp.stat.stat_utils import *
import multiprocessing as mp
from pyklip.kpp.utils.multiproc import *


def get_image_stat_map_perPixMasking(image,
                                     image_without_planet=None,
                                     mask_radius = 7,
                                     IOWA = None,
                                     N = None,
                                     centroid = None,
                                     N_threads = None,
                                     Dr = 2,
                                     Dth = None,
                                     type = "SNR"):
    """
    Calculate the SNR, the standard deviation or the probability (tail distribution) of a given image on a per pixel
    basis, which means that for each pixel the standard deviation is calculated after masking its surroundings.

    Args:
        image: The image for which one wants the statistic.
        image_without_planet: Same as image but where real signal has been masked out. The
        image_without_planet: Same as image but where any real signal has been masked out. The code will use
                            "image_without_planet" to calculate the standard deviation or the PDF.
                            This can be a negatively derotated image in the context of ADI.
        mask_radius: (default=7pix) Radius of the mask used around the current pixel.
        IOWA: (IWA,OWA) inner working angle, outer working angle. It defines boundary to the zones in which the
                    statistic is calculated.
                    If None, kpp.utils.GPIimage.get_IOWA() is used.
        N: Defines the width of the ring by the number of pixels it has to include.
                The width of the annuli will therefore vary with sepration.
        centroid: (x_cen,y_cen) Define the center of the image.
                Default is x_cen = (nx-1)//2 ; y_cen = (ny-1)//2
        N_threads: Number of threads to be used. If None run sequentially.
        Dr: (default=2pix) Width of the annulus (in pixel).
        Dth: Define the angular size of a sector in degree (will apply for either Dr or N)
        type: Indicate the type of statistic to be calculated.
                    If "SNR" (default) simple stddev calculation and returns SNR.
                    If "stddev" returns the pure standard deviation map.
                    If "proba" triggers proba calculation with pdf fitting.

    Return:
        The statistic map for image.
    """
    if image_without_planet is None:
        image_without_planet = image
    ny,nx = image.shape

    if IOWA is None:
        IWA,OWA = get_IOWA(image_without_planet, centroid = centroid)
    else:
        IWA,OWA = IOWA

    if centroid is None :
        x_cen = (nx-1)//2 ; y_cen = (ny-1)//2
    else:
        x_cen, y_cen = centroid

    image_without_planet_mask = np.ones((ny,nx))
    image_without_planet_mask[np.where(np.isnan(image_without_planet))] = 0

    # Build the x and y coordinates grids
    x_grid, y_grid = np.meshgrid(np.arange(nx)-x_cen, np.arange(ny)-y_cen)
    # Calculate the radial distance of each pixel
    r_grid = abs(x_grid +y_grid*1j)
    th_grid = np.arctan2(x_grid,y_grid)

    image_noNans = np.where(np.isfinite(image)*(r_grid>IWA)*(r_grid<OWA))

    stat_map = np.zeros(image.shape) + np.nan
    if N_threads is None:
        N_threads = mp.cpu_count()

    N_pix = image_noNans[0].size
    chunk_size = N_pix//N_threads
        
    if N_threads != -1 and chunk_size :
        try:
            pool = NoDaemonPool(processes=N_threads)
        except:
            pool = mp.Pool(processes=N_threads)

        N_chunks = N_pix//chunk_size

        # Shuffle the list of indices such that a thread doesn't end up with only the outer most pixel (where the number
        # of pixels in the pdf is much bigger which make it a lot slower compared to his brothers)
        image_noNans_rows = copy(image_noNans[0])
        image_noNans_cols = copy(image_noNans[1])
        rng_state = np.random.get_state()
        np.random.shuffle(image_noNans_rows)
        np.random.set_state(rng_state)
        np.random.shuffle(image_noNans_cols)

        # Get the chunks
        chunks_row_indices = []
        chunks_col_indices = []
        for k in range(N_chunks-1):
            chunks_row_indices.append(image_noNans_rows[(k*chunk_size):((k+1)*chunk_size)])
            chunks_col_indices.append(image_noNans_cols[(k*chunk_size):((k+1)*chunk_size)])
        chunks_row_indices.append(image_noNans_rows[((N_chunks-1)*chunk_size):N_pix])
        chunks_col_indices.append(image_noNans_cols[((N_chunks-1)*chunk_size):N_pix])

        outputs_list = \
            pool.map(get_image_stat_map_perPixMasking_threadTask_star,
                       zip(chunks_row_indices,
                       chunks_col_indices,
                       itertools.repeat(image),
                       itertools.repeat(image_without_planet),
                       itertools.repeat(x_grid),
                       itertools.repeat(y_grid),
                       itertools.repeat(N),
                       itertools.repeat(mask_radius),
                       itertools.repeat(Dr),
                       itertools.repeat(Dth),
                       itertools.repeat(type)))

        for row_indices,col_indices,out in zip(chunks_row_indices,chunks_col_indices,outputs_list):
            stat_map[(row_indices,col_indices)] = out
        pool.close()

    else:
        stat_map[image_noNans] = get_image_stat_map_perPixMasking_threadTask(image_noNans[0],
                                                                            image_noNans[1],
                                                                            image,
                                                                            image_without_planet,
                                                                            x_grid,y_grid,
                                                                            N,
                                                                            mask_radius,
                                                                            Dr = Dr,
                                                                            Dth = Dth,
                                                                            type = type)
        
    if type == "proba":
        return -np.log10(stat_map)
    else:
        return stat_map

def get_image_stat_map_perPixMasking_threadTask_star(params):
    """
    Convert `f([1,2])` to `f(1,2)` call.
    It allows one to call get_image_probability_map_perPixMasking_threadTask() with a tuple of parameters.
    """
    return get_image_stat_map_perPixMasking_threadTask(*params)

def get_image_stat_map_perPixMasking_threadTask(row_indices,
                                               col_indices,
                                               image,
                                               image_without_planet,
                                               x_grid,
                                               y_grid,
                                               N,
                                               mask_radius,
                                               Dr = 2,
                                               Dth = None,
                                               type = "SNR"):
    """
    Helper function for get_image_stat_map_perPixMasking().
    Calculate the SNR on the subset of pixels defined by row_indices,col_indices.

    Calculate the SNR, the standard deviation or the probability (tail distribution) of a subset of an image on a per
    pixel basis, which means that for each pixel the standard deviation is calculated after masking its surroundings.

    This function is used to ease parallelization.

    Args:
        row_indices: The row indices of images for which we want the statistic.
        col_indices: The column indices of images for which we want the statistic.
        image: The image or cubes for which one wants the statistic.
        image_without_planet: Same as image but where real signal has been masked out. The code will actually use
                                    map to calculate the standard deviation or the density function.
        mask_radius: Radius of the mask used around the current pixel when use_mask_per_pixel = True.
        IOWA: (IWA,OWA) inner working angle, outer working angle. It defines boundary to the zones in which the
                    statistic is calculated.
        N: Defines the width of the ring by the number of pixels it has to include.
                The width of the annuli will therefore vary with sepration.
        centroid: Define the cente rof the image. Default is x_cen = (nx-1)//2 ; y_cen = (ny-1)//2
        mute: Won't print any logs.
        N_threads: Number of threads to be used. If None run sequentially.
        Dr: (default=2pix) Defines the width of the ring (in pixel). N is then ignored.
        Dth: Define the angular size of a sector in degree (will apply for either Dr or N)
        type: Indicate the type of statistic to be calculated.
                    If "SNR" (default) simple stddev calculation and returns SNR.
                    If "stddev" returns the pure standard deviation map.
                    If "proba" triggers proba calculation with pdf fitting.

    Return: The statistic map for image.
    """

    ny,nx = image.shape

    image_without_planet_mask = np.ones((ny,nx))
    image_without_planet_mask[np.where(np.isnan(image_without_planet))] = 0

    # Calculate the radial distance of each pixel
    r_grid = abs(x_grid +y_grid*1j)
    th_grid = np.arctan2(x_grid,y_grid)
    if Dth != None:
        Dth_rad = Dth/180.*np.pi

    N_it = row_indices.size
    stat_map = np.zeros((N_it)) + np.nan
    #stdout.write("\r%d" % 0)
    for id,k,l in zip(range(N_it),row_indices,col_indices):
        if 1:#k == 109 and l == 135:
            #stdout.write("\r{0}/{1}".format(id,N_it))
            #stdout.flush()

            x = x_grid[(k,l)]
            y = y_grid[(k,l)]
            #print(x,y)
            r = r_grid[(k,l)]
            th = th_grid[(k,l)]

            if Dr is None:
                dr = N/(4*np.pi*r)
                r_min,r_max = (r-dr, r+dr)

            else:
                r_min,r_max = (r-Dr/2., r+Dr/2.)

            if Dth is None:
                if N is None:
                    where_ring = np.where((r_min< r_grid) * (r_grid < r_max) * image_without_planet_mask)
                else:
                    N_ring = np.pi*(r_max**2-r_min**2)
                    Dth_rad = np.pi*(N/N_ring)
                    # print((N/N_ring),Dth_rad)
                    delta_th_grid = np.mod(th_grid - th +np.pi,2.*np.pi)-np.pi
                    where_ring = np.where((r_min< r_grid) * (r_grid < r_max) * image_without_planet_mask * \
                                        (abs(delta_th_grid)<Dth_rad))
            else:
                delta_th_grid = np.mod(th_grid - th +np.pi,2.*np.pi)-np.pi
                where_ring = np.where((r_min< r_grid) * (r_grid < r_max) * image_without_planet_mask * \
                                    (abs(delta_th_grid)<(Dth_rad*50./r)))

            where_ring_masked = np.where((((x_grid[where_ring]-x)**2 +(y_grid[where_ring]-y)**2) > mask_radius*mask_radius))

            # if k==100 and l==100:
            #     import matplotlib.pyplot as plt
            #     im_cpy = copy(image)
            #     im_cpy[(where_ring[0][where_ring_masked],where_ring[1][where_ring_masked])] = 1000
            #     plt.figure(1)
            #     plt.imshow(im_cpy)
            #     plt.show()
            # # print(where_ring)

            # print(where_ring_masked)
            data = image_without_planet[(where_ring[0][where_ring_masked],where_ring[1][where_ring_masked])]

            if 0:#(k == 135 and l == 155) or (k == 162 and l == 165) or (k == 139 and l == 133) or (k == 165 and l == 132) :
                import matplotlib.pyplot as plt
                print(image[k,l])
                im_cpy = copy(image)
                im_cpy[(where_ring[0][where_ring_masked],where_ring[1][where_ring_masked])] = 1000
                #im_cpy[where_ring] = 1000
                plt.figure(1)
                plt.imshow(im_cpy)
                plt.show()

            if type == "proba":
                cdf_model, pdf_model, sampling, im_histo, center_bins  = get_cdf_model(data)

                cdf_fit = interp1d(sampling,cdf_model,kind = "linear",bounds_error = False, fill_value=1.0)
                stat_map[id] = 1-cdf_fit(image[k,l])
            elif type == "SNR":
                stat_map[id] = image[k,l]/np.nanstd(data)
            elif type == "stddev":
                stat_map[id] = np.nanstd(data)

    return stat_map




import multiprocessing as mp
import ctypes

import numpy as np
import pyklip.spectra_management as spec
import os

from pyklip.fmlib.nofm import NoFM
import pyklip.fm as fm

from scipy import interpolate
from copy import copy

debug = False


class FMPlanetPSF(NoFM):
    """
    Forward models the PSF of the planet through KLIP. Returns the forward modelled planet PSF
    """
    def __init__(self, inputs_shape, numbasis, sep, pa, dflux, input_psfs, input_wvs, flux_conversion=None, spectrallib=None, spectrallib_units="flux", star_spt=None, refine_fit=False, field_dependent_correction=None):
        """
        Defining the planet to characterizae

        Args:
            inputs_shape: shape of the inputs numpy array. Typically (N, y, x)
            numbasis: 1d numpy array consisting of the number of basis vectors to use
            sep: separation of the planet
            pa: position angle of the planet
            dflux: guess for contrast of planet averaged across band w.r.t star
            input_psfs: the psf of the image. A numpy array with shape (wv, y, x)
            input_wvs: the wavelegnths that correspond to the input psfs
                (doesn't need to be tiled to match the dimension of the input data of the instrument class)
            flux_conversion: an array of length N to convert from contrast to DN for each frame. Units of DN/contrast. 
                             If None, assumes dflux is the ratio between the planet flux and tthe input_psfs flux
            spectrallib: if not None, a list of spectra based on input_wvs
            spectrallib_units: can be either "flux"" or "contrast". Flux units requires dividing by the flux of the star to get contrast 
            star_spt: star spectral type, if None default to some random one
            refine_fit: (NOT implemented) refine the separation and pa supplied
        """
        # allocate super class
        super(FMPlanetPSF, self).__init__(inputs_shape, numbasis)

        self.inputs_shape = inputs_shape
        self.numbasis = numbasis
        self.sep = sep
        self.pa = pa
        self.dflux = dflux
        self.field_dependent_correction = field_dependent_correction

        if spectrallib_units.lower() != "flux" and spectrallib_units.lower() != "contrast":
            raise ValueError("spectrallib_units needs to be either 'flux' or 'contrast', not {0}".format(spectrallib_units))

        # only need spectral info if not broadband
        numwvs = np.size(input_wvs)
        if numwvs  > 1:
            if spectrallib is not None:
                self.spectrallib = spectrallib
            else:
                spectra_folder = os.path.dirname(os.path.abspath(spec.__file__)) + os.sep + "spectra" + os.sep
                spectra_files = [spectra_folder + "t650g18nc.flx"]
                self.spectrallib = [spec.get_planet_spectrum(filename, input_wvs)[1] for filename in spectra_files]

            # TODO: calibrate to contrast units
            # calibrate spectra to DN
            if spectrallib_units.lower() == "flux":
                # need to divide by flux of the star to get contrast units
                self.spectrallib = [spectrum/(spec.get_star_spectrum(input_wvs, star_type=star_spt)[1]) for spectrum in self.spectrallib]
            self.spectrallib = [spectrum/np.mean(spectrum) for spectrum in self.spectrallib]
        else:
            self.spectrallib = [np.array([1])]

        self.input_psfs = input_psfs
        self.input_psfs_wvs = input_wvs

        if flux_conversion is None:
            flux_conversion = np.ones(inputs_shape[0])
        self.flux_conversion = flux_conversion

        self.psf_centx_notscaled = {}
        self.psf_centy_notscaled = {}

        numwv,ny_psf,nx_psf =  self.input_psfs.shape
        x_psf_grid, y_psf_grid = np.meshgrid(np.arange(nx_psf * 1.) - nx_psf//2, np.arange(ny_psf * 1.) - ny_psf//2)
        psfs_func_list = []
        for wv_index in range(numwv):
            model_psf = self.input_psfs[wv_index, :, :] #* self.flux_conversion * self.spectrallib[0][wv_index] * self.dflux
            #psfs_interp_model_list.append(interpolate.bisplrep(x_psf_grid,y_psf_grid,model_psf))
            #psfs_interp_model_list.append(interpolate.SmoothBivariateSpline(x_psf_grid.ravel(),y_psf_grid.ravel(),model_psf.ravel()))
            psfs_func_list.append(interpolate.LSQBivariateSpline(x_psf_grid.ravel(),y_psf_grid.ravel(),model_psf.ravel(),x_psf_grid[0,0:nx_psf-1]+0.5,y_psf_grid[0:ny_psf-1,0]+0.5))
            #psfs_interp_model_list.append(interpolate.interp2d(x_psf_grid,y_psf_grid,model_psf,kind="cubic",bounds_error=False,fill_value=0.0))
            #psfs_interp_model_list.append(interpolate.Rbf(x_psf_grid,y_psf_grid,model_psf,function="gaussian"))

            if 0:
                import matplotlib.pylab as plt
                #print(x_psf_grid.shape)
                #print(psfs_interp_model_list[wv_index](x_psf_grid.ravel(),y_psf_grid.ravel()).shape)
                a = psfs_func_list[wv_index](x_psf_grid[0,:],y_psf_grid[:,0]).transpose()
                plt.figure()
                plt.subplot(1,3,1)
                plt.imshow(a,interpolation="nearest")
                plt.colorbar()
                ##plt.imshow(psfs_interp_model_list[wv_index](np.linspace(-10,10,500),np.linspace(-10,10,500)),interpolation="nearest")
                plt.subplot(1,3,2)
                plt.imshow(self.input_psfs[wv_index, :, :],interpolation="nearest")
                plt.colorbar()
                plt.subplot(1,3,3)
                plt.imshow(abs(self.input_psfs[wv_index, :, :]-a),interpolation="nearest")
                plt.colorbar()
                plt.show()

        self.psfs_func_list = psfs_func_list


    def alloc_fmout(self, output_img_shape):
        """Allocates shared memory for the output of the shared memory


        Args:
            output_img_shape: shape of output image (usually N,y,x,b)

        Returns:
            fmout: mp.array to store FM data in
            fmout_shape: shape of FM data array

        """
        fmout_size = int(np.prod(output_img_shape))
        fmout = mp.Array(self.data_type, fmout_size)
        fmout_shape = output_img_shape

        return fmout, fmout_shape


    def alloc_perturbmag(self, output_img_shape, numbasis):
        """
        Allocates shared memory to store the fractional magnitude of the linear KLIP perturbation
        Stores a number for each frame = max(oversub + selfsub)/std(PCA(image))

        Args:
            output_img_shape: shape of output image (usually N,y,x,b)
            numbasis: array/list of number of KL basis cutoffs requested

        Returns:
            perturbmag: mp.array to store linaer perturbation magnitude
            perturbmag_shape: shape of linear perturbation magnitude

        """
        perturbmag_shape = (output_img_shape[0], np.size(numbasis))
        perturbmag = mp.Array(self.data_type, int(np.prod(perturbmag_shape)))

        return perturbmag, perturbmag_shape


    def generate_models(self, input_img_shape, section_ind, pas, wvs, radstart, radend, phistart, phiend, padding, ref_center, parang, ref_wv, flipx):
        """
        Generate model PSFs at the correct location of this segment for each image denoated by its wv and parallactic angle

        Args:
            pas: array of N parallactic angles corresponding to N images [degrees]
            wvs: array of N wavelengths of those images
            radstart: radius of start of segment
            radend: radius of end of segment
            phistart: azimuthal start of segment [radians]
            phiend: azimuthal end of segment [radians]
            padding: amount of padding on each side of sector
            ref_center: center of image
            parang: parallactic angle of input image [DEGREES]
            ref_wv: wavelength of science image
            flipx: if True, flip x coordinate in final image

        Return:
            models: array of size (N, p) where p is the number of pixels in the segment
        """
        # create a blank canvas the same size as the input image centered at the psf center
        nx = input_img_shape[1] # x dimension of input image
        ny = input_img_shape[0] # y dimension of input image
        x_grid, y_grid = np.meshgrid(np.arange(nx * 1.)-ref_center[0], np.arange(ny * 1.)-ref_center[1])

        # retrieve wavelength, x and y dimensions of input psf
        numwv, ny_psf, nx_psf =  self.input_psfs.shape

        # create stamp for instrumental psf
        psf_lower = int(np.floor(ny_psf/2.0))    # lower bound
        psf_upper = int(np.ceil(ny_psf/2.0))     # upper bound
        psf_left = int(np.floor(nx_psf/2.0))   # left bound
        psf_right = int(np.ceil(nx_psf/2.0))    # right bound

        # a blank img array to write model PSFs into
        whiteboard = np.zeros((ny,nx))
        if debug:
            canvases = []
        models = []

        for pa, wv in zip(pas, wvs):

            # grab PSF given wavelength
            wv_index = spec.find_nearest(self.input_psfs_wvs,wv)[1]

            # find center of psf
            # to reduce calculation of sin and cos, see if it has already been calculated before
            if pa not in self.psf_centx_notscaled:
                # flipx requires the opposite rotation
                sign = -1.
                if flipx:
                    sign = 1.
                self.psf_centx_notscaled[pa] = self.sep * np.cos(np.radians(90. - sign*self.pa - pa))
                self.psf_centy_notscaled[pa] = self.sep * np.sin(np.radians(90. - sign*self.pa - pa))
            psf_centx = (ref_wv/wv) * self.psf_centx_notscaled[pa]
            psf_centy = (ref_wv/wv) * self.psf_centy_notscaled[pa]

            # create a coordinate system for the image that is with respect to the model PSF
            # round to nearest pixel and add offset for center
            planet_centx = int(round(psf_centx + ref_center[0]))
            planet_centy = int(round(psf_centy + ref_center[1]))
            
            # recenter stamp coordinate system about the location of the planet
            stamp_left = planet_centx-psf_left
            stamp_right = planet_centx+psf_right
            stamp_lower = planet_centy-psf_lower
            stamp_upper = planet_centy+psf_upper

            # Save the stamp length and width as variables for slicing the image
            stamp_len = slice(stamp_lower, stamp_upper)
            stamp_width = slice(stamp_left,stamp_right)

            x_vec_stamp_centered = np.copy(x_grid[0, stamp_width]) - psf_centx
            y_vec_stamp_centered = np.copy(y_grid[stamp_len, 0]) - psf_centy
            # rescale to account for the align and scaling of the refernce PSFs
            # e.g. for longer wvs, the PSF has shrunk, so we need to shrink the coordinate system
            x_vec_stamp_centered /= (ref_wv/wv)
            y_vec_stamp_centered /= (ref_wv/wv)


            # use intepolation spline to generate a model PSF of planet and write to temp img
            whiteboard[stamp_len, stamp_width] = \
                    self.psfs_func_list[wv_index](x_vec_stamp_centered,y_vec_stamp_centered).transpose()
            
            # if specified, make field dependent PSF correction
            if self.field_dependent_correction is not None:

                # find distance from center in x and y dimensions
                dx = x_grid[stamp_len, stamp_width]
                dy = y_grid[stamp_len, stamp_width]
                whiteboard[stamp_len, stamp_width] = \
                        self.field_dependent_correction(whiteboard[stamp_len, stamp_width], dx, dy)
    
            # write model img to output (segment is collapsed in x/y so need to reshape)
            whiteboard.shape = [input_img_shape[0] * input_img_shape[1]]
            segment_with_model = copy(whiteboard[section_ind])
            whiteboard.shape = [input_img_shape[0],input_img_shape[1]]

           

            models.append(segment_with_model)

            # clean whiteboard
            whiteboard[stamp_len, stamp_width] = 0

        


        return np.array(models)




    def fm_from_eigen(self, klmodes=None, evals=None, evecs=None, input_img_shape=None, input_img_num=None, ref_psfs_indicies=None, section_ind=None,section_ind_nopadding=None, aligned_imgs=None, pas=None,
                     wvs=None, radstart=None, radend=None, phistart=None, phiend=None, padding=None,IOWA = None, ref_center=None,
                     parang=None, ref_wv=None, numbasis=None, fmout=None, perturbmag=None, klipped=None, covar_files=None, flipx=True, **kwargs):
        """
        Generate forward models using the KL modes, eigenvectors, and eigenvectors from KLIP. Calls fm.py functions to
        perform the forward modelling

        Args:
            klmodes: unpertrubed KL modes
            evals: eigenvalues of the covariance matrix that generated the KL modes in ascending order
                   (lambda_0 is the 0 index) (shape of [nummaxKL])
            evecs: corresponding eigenvectors (shape of [p, nummaxKL])
            input_image_shape: 2-D shape of inpt images ([ysize, xsize])
            input_img_num: index of sciece frame
            ref_psfs_indicies: array of indicies for each reference PSF
            section_ind: array indicies into the 2-D x-y image that correspond to this section.
                         Note needs be called as section_ind[0]
            pas: array of N parallactic angles corresponding to N reference images [degrees]
            wvs: array of N wavelengths of those referebce images
            radstart: radius of start of segment
            radend: radius of end of segment
            phistart: azimuthal start of segment [radians]
            phiend: azimuthal end of segment [radians]
            padding: amount of padding on each side of sector
            IOWA: tuple (IWA,OWA) where IWA = Inner working angle and OWA = Outer working angle both in pixels.
                It defines the separation interva in which klip will be run.
            ref_center: center of image
            numbasis: array of KL basis cutoffs
            parang: parallactic angle of input image [DEGREES]
            ref_wv: wavelength of science image
            fmout: numpy output array for FM output. Shape is (N, y, x, b)
            perturbmag: numpy output for size of linear perturbation. Shape is (N, b)
            klipped: PSF subtracted image. Shape of ( size(section), b)
            kwargs: any other variables that we don't use but are part of the input
        """
        sci = aligned_imgs[input_img_num, section_ind[0]]
        refs = aligned_imgs[ref_psfs_indicies, :]
        refs = refs[:, section_ind[0]]


        # generate models for the PSF of the science image
        model_sci = self.generate_models(input_img_shape, section_ind, [parang], [ref_wv], radstart, radend, phistart, phiend, padding, ref_center, parang, ref_wv, flipx)[0]
        model_sci *= self.flux_conversion[input_img_num] * self.spectrallib[0][np.where(self.input_psfs_wvs == ref_wv)] * self.dflux

        # generate models of the PSF for each reference segments. Output is of shape (N, pix_in_segment)
        models_ref = self.generate_models(input_img_shape, section_ind, pas, wvs, radstart, radend, phistart, phiend, padding, ref_center, parang, ref_wv, flipx)

        # Calculate the spectra to determine the flux of each model reference PSF
        total_imgs = np.size(self.flux_conversion)
        num_wvs = self.spectrallib[0].shape[0]
        input_spectrum = self.flux_conversion[:num_wvs] * self.spectrallib[0] * self.dflux
        input_spectrum = np.ravel(np.tile(input_spectrum,(1, total_imgs//num_wvs)))
        input_spectrum = input_spectrum[ref_psfs_indicies]
        models_ref = models_ref * input_spectrum[:, None]

        # using original Kl modes and reference models, compute the perturbed KL modes (spectra is already in models)
        # also grab the pertrubed (C_AS) covariance matrix to compute its eigenvalues
        delta_KL, covar_perturb = fm.perturb_specIncluded(evals, evecs, klmodes, refs, models_ref, return_perturb_covar=True)

        # calculate postklip_psf using delta_KL
        postklip_psf, oversubtraction, selfsubtraction = fm.calculate_fm(delta_KL, klmodes, numbasis, sci, model_sci, inputflux=None)

        # calculate validity of linear perturbation on KLIP modes
        pca_img = (sci - np.nanmean(sci))[:, None] - klipped # shape of ( size(section), b)
        perturb_frac = np.nanmax(np.abs(oversubtraction + selfsubtraction), axis=1)/np.nanstd(pca_img, axis=0) # array of b
        this_validity = fm.calculate_validity(covar_perturb, models_ref, numbasis, evals, covar_files, evecs, klmodes, delta_KL) # array of b
        #this_validity = fm.calculate_validity2(evals, models_ref, numbasis) # array of b
        perturbmag[input_img_num] = this_validity

        fmout_shape = fmout.shape

        # nan the same pixels as the klipped image
        klipped_nans = np.where(np.isnan(klipped))
        postklip_psf[:, klipped_nans[0]] = np.nan

        # write forward modelled PSF to fmout (as output)
        # need to derotate the image in this step
        for thisnumbasisindex in range(np.size(numbasis)):
                fm._save_rotated_section(input_img_shape, postklip_psf[thisnumbasisindex], section_ind,
                                 fmout[input_img_num, :, :,thisnumbasisindex], None, parang,
                                 radstart, radend, phistart, phiend, padding,IOWA, ref_center, flipx=flipx)



    def cleanup_fmout(self, fmout):
        """
        After running KLIP-FM, we need to reshape fmout so that the numKL dimension is the first one and not the last

        Args:
            fmout: numpy array of ouput of FM

        Return:
            fmout: same but cleaned up if necessary
        """

        # Let's reshape the output images
        # move number of KLIP modes as leading axis (i.e. move from shape (N,y,x,b) to (b,N,y,x)
        dims = fmout.shape
        fmout = np.rollaxis(fmout.reshape((dims[0], dims[1], dims[2], dims[3])), 3)
        return fmout

    def save_fmout(self, dataset, fmout, outputdir, fileprefix, numbasis, klipparams=None, calibrate_flux=False,
                   spectrum=None, pixel_weights=1):
        """
        Saves the FM planet PSFs to disk. Saves both a KL mode cube and spectral cubes

        Args:
            dataset: Instruments.Data instance. Will use its dataset.savedata() function to save data
            fmout: the fmout data passed from fm.klip_parallelized which is passed as the output of cleanup_fmout
            outputdir: output directory
            fileprefix: the fileprefix to prepend the file name
            numbasis: KL mode cutoffs used
            klipparams: string with KLIP-FM parameters
            calibrate_flux: if True, flux calibrate the data in the same way as the klipped data
            spectrum: if not None, spectrum to weight the data by. Length same as dataset.wvs
            pixel_weights: weights for each pixel for weighted mean. Leave this as a single number for simple mean
        """
        weighted = len(np.shape(pixel_weights)) > 1
        numwvs = dataset.numwvs
        fmout_spec = fmout.reshape([fmout.shape[0], fmout.shape[1]//numwvs, numwvs,
                                            fmout.shape[2], fmout.shape[3]]) # (b, N_cube, wvs, y, x) 5-D cube

        # collapse in time and wavelength to examine KL modes
        if spectrum is None:
            KLmode_cube = np.nanmean(pixel_weights * fmout_spec, axis=(1,2))
            if weighted:
                # if the pixel weights aren't just 1 (i.e., weighted case), we need to normalize for that
                KLmode_cube /= np.nanmean(pixel_weights, axis=(1,2))
        else:
            #do the mean combine by weighting by the spectrum
            spectra_template = spectrum.reshape(fmout_spec.shape[1:3]) #make same shape as dataset.output
            KLmode_cube = np.nanmean(pixel_weights * fmout_spec * spectra_template[None,:,:,None,None], axis=(1,2))\
                            / np.nanmean(spectra_template[None, :, :, None, None] * pixel_weights, axis = (1, 2))

        # save FM location into header
        more_keywords = {'fm_sep': self.sep, 'fm_pa': self.pa}

        # broadband flux calibration for KL mode cube
        if calibrate_flux:
            KLmode_cube = dataset.calibrate_output(KLmode_cube, spectral=False)
        dataset.savedata(outputdir + '/' + fileprefix + "-fmpsf-KLmodes-all.fits", KLmode_cube,
                         klipparams=klipparams.format(numbasis=str(numbasis)), filetype="KL Mode Cube",
                         zaxis=numbasis, more_keywords=more_keywords)

        # if there is more than one wavelength, save spectral cubes
        if dataset.numwvs > 1:
            
            KLmode_spectral_cubes = np.nanmean(pixel_weights * fmout_spec, axis=1)
            if weighted:
                # if the pixel weights aren't just 1 (i.e., weighted case), we need to normalize for that. 
                KLmode_spectral_cubes /= np.nanmean(pixel_weights, axis=1)
            
            for KLcutoff, spectral_cube in zip(numbasis, KLmode_spectral_cubes):
                # calibrate spectral cube if needed
                if calibrate_flux:
                    spectral_cube = dataset.calibrate_output(spectral_cube, spectral=True)
                dataset.savedata(outputdir + '/' + fileprefix + "-fmpsf-KL{0}-speccube.fits".format(KLcutoff),
                                 spectral_cube, klipparams=klipparams.format(numbasis=KLcutoff),
                                 filetype="PSF Subtracted Spectral Cube", more_keywords=more_keywords)





__author__ = 'JB'

import multiprocessing as mp
import multiprocessing.pool as mpPool

class NoDaemonProcess(mp.Process):
    # make 'daemon' attribute always return False
    def _get_daemon(self):
        return False
    def _set_daemon(self, value):
        pass
    daemon = property(_get_daemon, _set_daemon)

# We sub-class multiprocessing.pool.Pool instead of multiprocessing.Pool
# because the latter is only a wrapper function, not a proper class.
class NoDaemonPool(mpPool.Pool):
    Process = NoDaemonProcess


#!/usr/bin/env python

"""
Observational SNR and exposure time calculator
Jonathan Aguilar
Nov. 15, 2013
"""
from __future__ import division

import sys
import os
import warnings
import glob

import argparse
#for handling different python versions
if sys.version_info < (3,0):
    import ConfigParser
else:
    import configparser as ConfigParser

import matplotlib as mpl
try:
    mpl.use('TkAgg')
except UserWarning:
    pass

import time

import numpy as np
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
from matplotlib.colors import LogNorm, Normalize
import matplotlib.ticker as mticker
from matplotlib.collections import PatchCollection
from matplotlib.patches import CirclePolygon

if sys.version_info[0] < 3:
    import Tkinter as Tk
    import tkFileDialog
else:
    import tkinter as Tk
    import filedialog

import tkFont

from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg, NavigationToolbar2TkAgg

from astropy.io import fits
from astropy import units

dnah_spot_directory = '/data/p1640/data/users/spot_positions/jonathan/'


# some text formatting
mpl.rcParams["font.size"]=16
mpl.rcParams["text.color"]='black'
mpl.rcParams["axes.labelcolor"]='black'
mpl.rcParams["xtick.color"]='black'
mpl.rcParams["ytick.color"]='black'



class CubeChecker:

    def __init__(self, master, fitsfiles,
                 spot_mode=False, spot_path=dnah_spot_directory):
        master.wm_title("P1640 Cube Inspector")
        self.customFont = tkFont.Font(family="Helvetica", size=12)

        # file list window
        frame = Tk.Frame(master)
        frame.grid()
        # cube display window
        frame_cube = Tk.Toplevel()
        frame_cube.title("Current cube")
        frame_cube.grid()
        # good cubes window
        self.frame_good_cubes = Tk.Toplevel()
        self.frame_good_cubes.title("Good cubes")
        self.frame_good_cubes.grid()

        self.spot_mode = spot_mode
        self.spot_path = spot_path
        self.fitsfiles = sorted([os.path.abspath(i) for i in fitsfiles])
        try:
            self.seeing = [fits.getval(ff,'SEEING') for ff in self.fitsfiles]
            self.airmass = [fits.getval(ff,'INIT_AM') for ff in self.fitsfiles]
        except KeyError:
            self.seeing = np.ones(len(self.fitsfiles))*10
            self.airmass = np.ones(len(self.fitsfiles))*10
        self.quit_button = Tk.Button(frame, text="Quit", font=self.customFont, fg='black', underline=0,
                                     command=frame.quit)

        # initialize the cube
        self.current_file_index = 0
        self.current_cube_path = Tk.StringVar()
        self.current_cube_path.set(self.fitsfiles[self.current_file_index])
        self.current_cube = fits.getdata(self.current_cube_path.get())
        self.current_header = fits.getheader(self.current_cube_path.get())


        # radio buttons to select linear or log scaling
        self.scaling_group = Tk.LabelFrame(frame_cube, text="Image Scaling", font=self.customFont)
        self.image_scaling = Tk.StringVar()
        self.image_scaling.set("linear")
        self.scale_select_radio = []
        self.scale_select_radio.append(Tk.Radiobutton(self.scaling_group, text="Linear",
                                                      font=self.customFont,
                                                      variable = self.image_scaling,
                                                      value="linear",
                                                      command=self.update_cubeax_display))
        self.scale_select_radio.append(Tk.Radiobutton(self.scaling_group, text="Log",
                                                      font=self.customFont,
                                                      variable=self.image_scaling,
                                                      value="log",
                                                      command=self.update_cubeax_display))

        ### Cube slider and display
        self.f_cube = mpl.figure.Figure()
        self.canvas_cube = FigureCanvasTkAgg(self.f_cube, master=frame_cube)
#        self.canvas_cube.show()
        self.ax_cube = self.f_cube.add_subplot(111)
        #self.ax_cube.xaxis.set_ticks_position("bottom")
        #self.ax_cube.set_xticklabels(self.ax_cube.get_xticklabels(), va='bottom')
        self.s_cube = Tk.Scale(frame_cube,
                               from_=0,to=len(self.current_cube)-1,
                               label='Channel',
                               orient='horizontal',
                               troughcolor='blue',
                               font=self.customFont,
                               length=400,
                               resolution=1,
                               command=self.update_cubeax_display,
                               repeatdelay=300,
                               repeatinterval=250,
                               tickinterval=5)
        self.s_cube.set(25)
        self.play_cube_var = Tk.IntVar()
        self.play_cube_var.set(0)
        self.play_cube_button = Tk.Checkbutton(frame_cube,
                                               text="Auto scroll",
                                               font=self.customFont,
                                               variable=self.play_cube_var,
                                               onvalue=1,
                                               offvalue=0,
                                               command=self.toggle_autoscroll)
                                          
        # plot seeing and airmass
        self.f_seeing = mpl.figure.Figure(figsize=(4,3))
        self.f_airmass = mpl.figure.Figure(figsize=(4,3))
        self.canvas_seeing = FigureCanvasTkAgg(self.f_seeing, master=frame)
        self.canvas_seeing.show()
        self.ax_seeing = self.f_seeing.add_subplot(111)
        self.canvas_airmass = FigureCanvasTkAgg(self.f_airmass, master=frame)
        self.canvas_airmass.show()
        self.ax_airmass = self.f_airmass.add_subplot(111)
   
        # Make a frame to organize the file selection tools
        self.file_lists = Tk.LabelFrame(frame, text="Cube list", font=self.customFont, labelanchor=Tk.N)
        # show the data directory
        self.data_dir = Tk.Label(frame,#self.file_lists,
                                 text='Path:\t'+os.path.dirname(os.path.commonprefix(self.fitsfiles)),
                                 justify=Tk.LEFT,
                                 #wraplength=300,
                                 anchor=Tk.NW,
                                 font=self.customFont)
        # radio buttons to select a fits file, and check boxes to select good files
        self.current_file_group = Tk.LabelFrame(self.file_lists, text="Select", font=self.customFont, padx=5, pady=5)
        self.good_file_group = Tk.LabelFrame(self.file_lists, text="Keep?", font=self.customFont, padx=0, pady=5)
        self.file_select_radio = []
        self.file_select_check = []
        self.good_files = []
        for i,ff in enumerate(self.fitsfiles):
            self.file_select_radio.append(Tk.Radiobutton(self.current_file_group,
                                                         anchor=Tk.W,
                                                         text=os.path.basename(ff),
                                                         font=self.customFont,
                                                         variable=self.current_cube_path,
                                                         value=ff,
                                                         indicatoron=0,
                                                         justify=Tk.LEFT,
                                                         offrelief=Tk.RIDGE,
                                                         command=self.load_selected_cube))
            gf_var = Tk.StringVar()
            self.file_select_check.append(Tk.Checkbutton(self.good_file_group,
                                                         text='',
                                                         font=self.customFont,
                                                         variable=gf_var,
                                                         onvalue=self.fitsfiles[i],
                                                         offvalue=""))
            self.good_files.append(gf_var)
        self.keep_button = Tk.Button(frame_cube, text="Keep?", font=self.customFont, fg="black", underline=0,
                                     command=self.toggle_check)
        self.next_button = Tk.Button(frame_cube, text="Next", font=self.customFont, fg="black", underline=0,
                                     command=self.next_cube)
        self.prev_button = Tk.Button(frame_cube, text="Prev", font=self.customFont, fg="black", underline=0,
                                     command=self.prev_cube)

        # bindings
        self.bind_buttons_to_frame(master)
        self.bind_buttons_to_frame(frame_cube)
        self.bind_buttons_to_frame(self.frame_good_cubes)
        
        # cube data fields
        self.cube_info_group = Tk.LabelFrame(frame_cube, text="Exposure Info", font=self.customFont)
        self.curr_exptime = Tk.StringVar()
        self.curr_seeing = Tk.StringVar()
        self.curr_airmass = Tk.StringVar()
        self.curr_minval = Tk.StringVar()
        self.curr_maxval = Tk.StringVar()
        # assign labels to these variables
        self.exptime_label = Tk.Label(self.cube_info_group,
                                      textvariable=self.curr_exptime, font=self.customFont)
        self.seeing_label = Tk.Label(self.cube_info_group,
                                     textvariable=self.curr_seeing, font=self.customFont)
        self.airmass_label = Tk.Label(self.cube_info_group,
                                      textvariable=self.curr_airmass, font=self.customFont)
        self.max_label = Tk.Label(self.cube_info_group,
                                  textvariable=self.curr_maxval, font=self.customFont)
        self.min_label = Tk.Label(self.cube_info_group,
                                  textvariable=self.curr_minval, font=self.customFont)

       
        # print good files button
        self.print_good_button = Tk.Button(frame, text="Print good files", font=self.customFont, underline=0,
                                           fg='black', command=self.print_good_cubes)
        self.wiki_format_flag = Tk.IntVar()
        self.wiki_format_flag.set(1)
        self.print_wiki_format = Tk.Checkbutton(frame, text="Output as wiki-pasteable list",
                                                font=self.customFont, variable=self.wiki_format_flag,
                                                fg='black')
        
        

        # File list window
        self.data_dir.grid(row=0,column=0, columnspan=4, padx=20)
        self.quit_button.grid(row=0, column=4, columnspan=2,  padx=20,pady=5, sticky=Tk.E)
        self.print_good_button.grid(row=1, column=0, columnspan=2, padx=20,pady=5, sticky=Tk.W)
        self.print_wiki_format.grid(row=2,column=0, padx=20, pady=2, sticky=Tk.NW)
        self.file_lists.grid(row=3,column=0,columnspan=2, rowspan=8, padx=20, sticky=Tk.NW)


        self.canvas_seeing.get_tk_widget().grid(row=1, column=3, rowspan=3, padx=20, pady=20)
        self.canvas_airmass.get_tk_widget().grid(row=4, column=3, rowspan=3, padx=20, pady=20)

        # file lists
        self.current_file_group.grid(row=0, column=0, rowspan=8, sticky=Tk.W)
        self.good_file_group.grid(row=0, column=1, rowspan=8)
        # file selection
        for i, fsr in enumerate(self.file_select_radio):
            fsr.grid(row=i,column=0)
        for i, fsc in enumerate(self.file_select_check):
            fsc.grid(row=i,column=1)

        # Current cube window
        self.canvas_cube.get_tk_widget().grid(row=1,column=1,rowspan=5,columnspan=5, padx=20)
        self.s_cube.grid(row=6,column=1,columnspan=5)

        self.scaling_group.grid(row=5,column=7)

        # self.play_cube_button.grid(row=7,column=0) # disabled for now because it doesn't work
        self.prev_button.grid(row=8,column=2)        
        self.keep_button.grid(row=8,column=3)
        self.next_button.grid(row=8,column=4)

        self.cube_info_group.grid(row=3,column=7, columnspan=2, sticky=Tk.W, padx=20)

        # current cube stats
        self.exptime_label.grid(row=1,column=0, sticky=Tk.W)
        self.seeing_label.grid(row=2,column=0, sticky=Tk.W)
        self.airmass_label.grid(row=3,column=0, sticky=Tk.W)
        self.max_label.grid(row=4,column=0, sticky=Tk.W)
        self.min_label.grid(row=5,column=0, sticky=Tk.W)
        
        # image scaling
        self.scale_select_radio[0].grid(row=1,column=1, sticky=Tk.W)
        self.scale_select_radio[1].grid(row=2,column=1,sticky=Tk.W)

        # initialize the whole widget based on the selected cube
        self.load_selected_cube()

    # key bindings
    def next_button_pushed(self, event):
        self.next_button.invoke()
    def prev_button_pushed(self, event):
        self.prev_button.invoke()
    def quit_button_pushed(self, event):
        self.quit_button.invoke()
    def keep_button_pushed(self, event):
        self.keep_button.invoke()
    def scroll_cube_left(self, event):
        if self.s_cube.get() == 0:
            self.s_cube.set(len(self.current_cube)-1)
        else:
            self.s_cube.set(np.max([self.s_cube.get()-1,0]))
    def scroll_cube_right(self, event):
        self.s_cube.set((self.s_cube.get()+1)%len(self.current_cube))
    def activate_printing(self, event):
        self.print_good_button.invoke()
        
    def bind_buttons_to_frame(self, frame_ref):
        """
        These are the buttons you want to be recognized by all the frames
        """
        frame_ref.bind("n", self.next_button_pushed)
        frame_ref.bind("p", self.prev_button_pushed)
        frame_ref.bind("k", self.keep_button_pushed)
        frame_ref.bind("q", self.quit_button_pushed)
        frame_ref.bind("<Left>", self.scroll_cube_left) 
        frame_ref.bind("<Right>", self.scroll_cube_right)
        frame_ref.bind("<Up>", self.prev_button_pushed) 
        frame_ref.bind("<Down>", self.next_button_pushed)
        frame_ref.bind("<Shift-P>", self.activate_printing)

        
    @property
    def current_cube(self):
        return self._current_cube
    @current_cube.setter
    def current_cube(self, newval):
        self._current_cube = newval

    #### Commands ####
    def load_selected_cube(self):
        """
        Read the highlighted fitsfile, get the datacube from it, and display it
        This function does the heavy lifting and calls the other update functions
          - check for spot mode
          - set the current file index
          - set the current file using the current file index
          - load the cube from the current file
        """
        # find the new file index
        self.current_file_index = self.fitsfiles.index(self.current_cube_path.get())
        # update the cube and header
        self.current_cube = fits.getdata(self.current_cube_path.get(), 0)
        if self.spot_mode:
            self.current_cube[:,90:160,90:160] = np.nan
        self.current_header = fits.getheader(self.current_cube_path.get(),0)
        # run update functions
        self.update_seeing_and_airmass()
        self.update_cubeax_display()
        self.update_cube_stats()
        
    def update_cube_stats(self):
        header = fits.getheader(self.fitsfiles[self.current_file_index])
        try:
            exptime = header['EXP_TIME']
            airmass = np.mean([header['INIT_AM'], header['FINL_AM']])
            seeing = header['SEEING']
        except KeyError:
            exptime = np.nan
            airmass = np.nan
            seeing = np.nan
        maxval = np.nanmax(self.current_cube)
        minval = np.nanmin(self.current_cube)
        """
        self.curr_exptime.set("Exposure time: {0:.3f}".format(self.current_header['EXP_TIME']))
        self.curr_seeing.set("Seeing: {0:>9.3f}".format(self.current_header['SEEING']))
        self.curr_airmass.set("Airmass: {0:>8.3f}".format(self.current_header['INIT_AM']))
        self.curr_maxval.set("Max val: {0:>8.1f}".format(np.nanmax(self.current_cube)))
        self.curr_minval.set("Min val: {0:>8.1f}".format(np.nanmin(self.current_cube)))
        """
        self.curr_exptime.set("Exposure time: {0:.3f}".format(exptime))
        self.curr_seeing.set("Seeing: {0:>9.3f}".format(seeing))
        self.curr_airmass.set("Airmass: {0:>8.3f}".format(airmass))
        self.curr_maxval.set("Max val: {0:>8.1f}".format(np.nanmax(self.current_cube)))
        self.curr_minval.set("Min val: {0:>8.1f}".format(np.nanmin(self.current_cube)))

    ## Spots ##
    def load_spot_files(self):
        """
        based on the current cube and the spot path, get the spot files
        """
        cubefile_name = os.path.splitext(os.path.basename(self.current_cube_path.get()))[0]
        files = sorted(glob.glob(os.path.join(self.spot_path, cubefile_name)+"*"))
        return files
    
    def draw_spots_on_cube(self):
        spot_files = self.load_spot_files()
        spots = [np.genfromtxt(f, delimiter=',') for f in spot_files]
        chan = self.s_cube.get()
        patches1 = [CirclePolygon(xy=spot[chan][::-1], radius=5,
                                  fill=False, alpha=1, ec='k', lw=2)
                    for spot in spots] # large circles centered on spot
        patches2 = [CirclePolygon(xy=spot[chan][::-1], radius=1,
                                  fill=True, alpha=0.3, ec='k', lw=2)
                    for spot in spots] # dots in location of spot
        patchcoll = PatchCollection(patches1+patches2, match_original=True)
        self.ax_cube.add_collection(patchcoll)
        self.f_cube.canvas.draw_idle()
        
    ## Cube selection shortcuts ##
    def toggle_check(self):
        self.file_select_check[self.current_file_index].toggle()
    def next_cube(self):
        # make sure you're in range
        if self.current_file_index+1 == len(self.file_select_radio):
            pass
        else:
            rad = self.file_select_radio[self.current_file_index+1]
            rad.select()
            rad.invoke()
    def prev_cube(self):
        # make sure you're in range
        if self.current_file_index == 0:
            pass
        else:
            rad = self.file_select_radio[self.current_file_index-1]
            rad.select()
            rad.invoke()

    ## Plotting ##
    def update_cubeax_display(self, event=None):
        img_norm = Normalize
        if self.image_scaling.get() == "log":
            img_norm = LogNorm
        self.ax_cube.clear()

        self.ax_cube.matshow(self.current_cube[self.s_cube.get()], origin='lower', norm=img_norm(),
                             cmap='cubehelix')
        self.ax_cube.xaxis.set_ticks_position("bottom")
        self.ax_cube.set_title(os.path.basename(self.current_cube_path.get()))
        if self.spot_mode is True:
            self.draw_spots_on_cube()
        self.f_cube.canvas.draw_idle()

    def toggle_autoscroll(self):
        len_cube = len(self.current_cube)
        while self.play_cube_var.get() == 1:
            self.s_cube.set((self.s_cube.get()+1)%len_cube)
            time.sleep(0.3) # maybe this needs to be longer

        
    def update_seeing_and_airmass(self):
        self.ax_seeing.clear()
        self.ax_airmass.clear()

        self.ax_seeing.plot(self.seeing,'r')
        self.ax_airmass.plot(self.airmass,'r')

        # axis styling
        self.ax_seeing.set_xlim(xmin=-0.5, xmax=len(self.seeing)+0.5)
        self.ax_seeing.set_yticks(np.sort(self.seeing)[[0,-1]])
        self.ax_seeing.axvline(self.current_file_index, c='k', ls='--')
        self.ax_seeing.axhline(self.seeing[self.current_file_index], c='k', ls='--')
        self.ax_seeing.yaxis.set_major_formatter(mticker.FormatStrFormatter('%.2f'))
        self.ax_seeing.set_title("seeing")
        self.ax_seeing.grid(True)
        self.f_seeing.tight_layout()
        self.f_seeing.canvas.draw_idle()

        self.ax_airmass.set_xlim(xmin=-0.5, xmax=len(self.airmass)+0.5)
        self.ax_airmass.set_yticks(np.sort(self.airmass)[[0,-1]])
        self.ax_airmass.axvline(self.current_file_index, c='k', ls='--')
        self.ax_airmass.axhline(self.airmass[self.current_file_index], c='k', ls='--')
        self.ax_airmass.yaxis.set_major_formatter(mticker.FormatStrFormatter('%.2f'))
        self.ax_airmass.set_title("airmass")
        self.ax_airmass.grid(True)
        self.ax_airmass.set_xlabel("cube #", size='large')
        self.f_airmass.tight_layout()
        self.f_airmass.canvas.draw_idle()

    ## Printing ##
    def print_good_cubes(self):
        good_cubes = [os.path.abspath(i.get()) for i in self.good_files if i.get()]
        # clear the good_cubes frame
        children = self.frame_good_cubes.winfo_children()
        if len(children) > 0:
            for child in children:
                child.destroy()
        
        try:
            max_length = max([len(i) for i in good_cubes])
        except ValueError:
            max_length=0
        width = max([max_length, 20])
        exptimevar = Tk.StringVar()
        exptimevar.set("0 min")
        exptime_label = Tk.Label(self.frame_good_cubes,
                                 textvariable=exptimevar,
                                 font=self.customFont)
        good_cubes_listbox = Tk.Listbox(self.frame_good_cubes,
                                        exportselection=1,
                                        height=len(good_cubes)+2,
                                        selectmode=Tk.EXTENDED,
                                        font=self.customFont,
                                        width=width + 5)
        
        if len(good_cubes) > 0:
            #print("Printing good cubes:")
            for i in good_cubes:
                #print i
                if self.wiki_format_flag.get() == 1:
                    i = "  * " + i
                good_cubes_listbox.insert(Tk.END, i)
            exptime = get_total_exposure_time(good_cubes)
            expmin = np.int(np.floor(exptime.value))
            expsec = exptime.decompose().value % 60
            exptimevar.set("Total exposure time: {0} min {1:0.0f} sec".format(expmin, expsec))
            #print(exptimestr)
        else:
            #print("No good cubes")
            good_cubes_listbox.insert(Tk.END, "No good cubes")
        exptime_label.grid(row=0,column=0)
        good_cubes_listbox.grid(row=1,column=0,rowspan=len(good_cubes)+3)
        self.bind_buttons_to_frame(self.frame_good_cubes)
        # after you push this button, raise the window
        self.frame_good_cubes.lift()

#### Helper functions ####



def get_total_exposure_time(fitsfiles, unit=units.minute):
    """
    Accept a list of fits files and return the total exposure time
    Input:
      fitsfiles: single fits file or list of files with keyword 'EXPTIME' in the header
      units: [minute] astropy.units unit for the output
    Output:
      tot_exp_time: the sum of the exposure times for each cube, in minutes
    """
    exptimes = np.array([fits.getval(f, 'EXP_TIME') for f in fitsfiles]) * units.second
    return np.sum(exptimes).to(unit)


# CLI argument handling
# Command-line option handling
# if switch --config is used, treat the following argument like the path to a config file
# if --config is not present, treat the following arguments like individual datacubes
class ConfigAction(argparse.Action):
    """
    Create a custom action to parse the command line arguments
    """
    def __init__(self, option_strings, dest, nargs=None, **kwargs):
        #if nargs is not None:
        #    raise ValueError("nargs not allowed")
        super(ConfigAction, self).__init__(option_strings, dest, **kwargs)
    def __call__(self, parser, namespace, values, option_string=None):
        configparser = ConfigParser.ConfigParser()
        configparser.read(values)
        # get the list of files
        filelist = configparser.get("Input","occulted_files").split()
        setattr(namespace, self.dest, filelist)
        try:
            setattr(namespace, 'spot_path', configparser.get("Spots","spot_file_path"))
        except:
            setattr(namespace, 'spot_path', None)

parser = argparse.ArgumentParser(prog="P1640_cube_checker.py",
                                 description='A utility to visually inspect P1640 cubes and spots')
#parser.add_argument('--help', action=usage)
group = parser.add_mutually_exclusive_group(required=True)
group.add_argument('--files', dest='files', nargs='*',
                    help='list of datacube files')
group.add_argument('--config', dest='files', nargs=1, action=ConfigAction,
                    help='config file containing fits files')
spotargs = parser.add_argument_group("Spot fitting", description="Optional arguments, use if you want to check spot locations")
spotargs.add_argument("--spots", dest='spots',action='store_true', default=False,
                      help="use this flag if you want to overplot spot positions")
spotargs.add_argument("--spot_path", dest='spot_path', action='store',
                       default=dnah_spot_directory,
                       help='directory where spot position files are stored')



        
if __name__ == "__main__":

    # some initializations
    spot_mode = False
    spot_directory = dnah_spot_directory
    
    parseobj = parser.parse_args(sys.argv[1:])
    try:
        fitsfiles = parseobj.files
        if parseobj.spots:
            spot_mode=True
            spot_directory = parseobj.spot_path
    except:
        sys.exit()

    try:
        # run the gui
        root = Tk.Tk()
        cubechecker = CubeChecker(root, fitsfiles, spot_mode, spot_directory)
        root.mainloop()
        root.destroy()
    except KeyboardInterrupt as e:
        pass



#!/usr/bin/env python

"""
Utilities specific to generating contrast curves
"""

import sys
import os
import glob
import re

#for handling different python versions
if sys.version_info < (3,0):
    import ConfigParser
else:
    import configparser as ConfigParser

import numpy as np
import pandas as pd
from scipy import interpolate

import matplotlib.pyplot as plt
import matplotlib as mpl
from matplotlib.colors import LogNorm


from astropy.io import fits
from photutils import aperture_photometry, CircularAperture

import pyklip.instruments.P1640_support.P1640cores

def calc_contrast_single_file(filename, core_info=None, chans='all'):
    """
    Calculate the radiall-averaged variance for a pyklip-reduced file for a single channel
    Input: 
        filename: full path to a pyklip-processed datacube
        core_info: pandas DataFrame with 'radius' and 'flux' columns
        chans: iterative type list of channels
    """
    # pick an arbitrary reference channel
    cube = fits.getdata(filename)
    numfiles = fits.getval(filename, 'DRPNFILE')
    starx = fits.getval(filename, 'PSFCENTX')
    stary = fits.getval(filename, 'PSFCENTY')
    # normalize flux for co-added cubes
    cube /= numfiles
    if chans == 'all':
        chans = range(len(cube))
    if np.size(chans) == 1:
        chans = [chans]

    # get radial coordinates
    yx = (np.indices(cube.shape[-2:]).T - [stary,starx]).T
    rad_coord = np.ravel(np.linalg.norm(yx, axis=0))
    radlims = (np.floor(np.nanmin(rad_coord)), np.ceil(np.nanmax(rad_coord)))
    radbins = np.linspace(radlims[0], radlims[1], 50)
    radii = radbins[:-1] + np.diff(radbins)
    
    # convolve cube with aperture
    convolved_ims = aperture_convolve_cube(cube[chans], core_info['radius'][chans])
    # divide by flux
    contrast_ims = (convolved_ims.T/np.array(core_info['flux'][chans])).T

    contrast_df = {}
    contrast_df['rad'] = radii
    for i, chan in enumerate(chans):
        radpix = pd.DataFrame(zip(rad_coord, np.ravel(contrast_ims[i])), 
                              columns=['rad', 'pixval'])
        radpix.sort(columns='rad', inplace=True)
        radpix.reset_index(inplace=True)
        # set up windowed variance
        std = []
        for j in range(len(radbins)-1):
            std.append(radpix[(radpix['rad'] >= radbins[j]) & (radpix['rad'] < radbins[j+1])]['pixval'].std())
        std = np.array(std)
        contrast_df['chan{0:02d}'.format(chan)] = std
        
    return pd.DataFrame(contrast_df)

def calc_contrast_multifile(core_files, datacube):
    """
    Assemble a median core PSF out of the list of core_files, and return
    a datacube scaled by the core flux
    """
    pass

def make_contrast_plot(contrast_map, title=None, 
                       contrast_range=None, 
                       plate_scale=19.2, pckwargs=None):
    """
    Plot contrast against separation and channel number.
    Input:
        contrast_map: Pandas DataFrame. See required columns below.
        name: plot title (preferably the file name corresponding to the contrast map)
        plate_scale (19.2 mas/pix): convert between pixels and mas
        contrast_range (None): tuple of upper and lower bounds for contrast plot. If none, min-max.
        pckwargs: dictionary of arguments that can be passed to plt.pcolor
    Returns:
       fig: plt.figure() object
    contrast_map: one column is titled 'rad', this is the separation in pixels
    The rest of the columns are titled 'chan##', where ## is the channel number.
    """
    # x- and y-axes
    channels = [col for col in contrast_map.columns if re.match('chan[0-9][0-9]', col) is not None]
    sep = contrast_map['rad']*plate_scale
    masked_data = np.ma.masked_array(contrast_map[channels].as_matrix(),
                                     mask=np.isnan(contrast_map[channels].as_matrix()))
    fig = plt.figure()
    if contrast_range is None: 
        contrast_range=(np.nanmin(masked_data), np.nanmax(masked_data))
    ax = plt.pcolor(sep, np.arange(len(channels)),
                    masked_data.T, 
                    cmap = 'cubehelix',
                    norm=LogNorm(vmin=contrast_range[0], 
                                 vmax=contrast_range[1]))
    plt.colorbar()
    plt.xlim(0, 3000)
    plt.ylim(0, len(channels)-1)
    plt.grid(True, which='both')
    plt.xlabel("Separation [mas]", size='x-large')
    plt.ylabel("Channel", size='x-large')
    if title is not None:
        plt.title(title, size='xx-large')
    fig.tight_layout()
    return fig    

def make_contrast_summary_plot(contrast_map_dict, chan='chan23', title=None, plate_scale=19.2, kwargs=None):
    """
    Plot the mean, min, and max contrast in the given channel for all the reductions
    """
    # select the diagnostic channel
    chandata = pd.concat([pd.DataFrame(cm[['rad',chan]]) for cm in contrast_map_dict.values()])
    # put the values at different radii together
    grouped = chandata.groupby('rad')
    rad = np.array(sorted(grouped.groups.keys()))*plate_scale
    fig = plt.figure()
    plt.fill_between(rad, grouped.min()[chan], grouped.max()[chan], alpha=0.5, label='min-max')
    plt.plot(rad, grouped.median()[chan], 'k-', lw=2, label='median')
    plt.yscale('log')
    plt.xlabel("Separation [mas]")
    plt.ylabel("Contrast")
    plt.grid(True, which='major')
    if title is None:
        title = chan + ' summary'
    plt.title(title, size='x-large')
    plt.legend(numpoints=1)
    fig.tight_layout()
    return fig



################################
# Run as script
################################
if __name__ == "__main__":
    """Run with -s for saving figures"""
    # Get star info from config file
    config = sys.argv[1]
    SaveFlag = False
    try:
        if sys.argv[2] == '-s':
            SaveFlag = True
    except IndexError:
        pass
    
    cfp = ConfigParser.SafeConfigParser()
    cfp.read(config)

    StarName = cfp.get("Contrast","StarName")
    ObsDate =  cfp.get("Contrast","ObsDate")
    print("Star: {star}\nEpoch: {epoch}".format(star=StarName,epoch=ObsDate))

    # validate input paths
    reduced_files = eval(cfp.get("Contrast","reduced_file_search_command"))
    # if you messed up the reduced file path, you get to try again
    while len(reduced_files) == 0:
        print("No reduced files found with: {0}".format(cfp.get("Contrast","reduced_file_search_command")))
        reduced_files = eval(raw_input("Fix command: "))
        if len(reduced_files) >= 0: print("{0} reduced files found, good job. Moving on...\n".format(len(reduced_files)))
    
    # STAR FLUX
    # prepare core files
    core_files = cfp.get("Contrast","core_files").split()
    core_hdus = [fits.open(f) for f in core_files]
    core_cubes = [hdulist[0].data for hdulist in core_hdus]
    for core in core_cubes: 
        core[core == 0] = np.nan
    print("Determining core fluxes and radii")
    core_info_all = [P1640cores.get_encircled_energy_cube(c) for c in core_cubes]
    core_info = P1640cores.combine_multiple_cores(core_info_all)
    print("    ...finished.\n")
    
    # POST-PROCESSING FLUX
    contrast_maps = {}
    channels = 'all'
    print("Generating contrast maps")
    for i,ff in enumerate(reduced_files):
        fname = os.path.basename(ff)
        print("\t{n:2d}/{tot}\t{ff}".format(n=i,tot=len(reduced_files),ff=fname))
        contrast_maps[fname] = calc_contrast_single_file(ff, core_info, chans=channels)
        print("Processing complete, contrast maps stored in variable 'contrast_maps'")

    plate_scale = np.float(cfp.get("Contrast","plate_scale"))
    if SaveFlag is True:
        figure_path = cfp.get("Contrast","figure_path")
        sumchan = 'chan23'
        fig = make_contrast_summary_plot(contrast_maps,
                                         title=StarName+' '+ObsDate+'\n'+sumchan+' '+'summary',
                                         plate_scale=plate_scale)
        fig.savefig(os.path.join(figure_path, StarName+'_'+ObsDate+'-summary.png'))
        plt.close(fig.number)
        for fname, cm in contrast_maps.iteritems():
            name = os.path.splitext(os.path.basename(fname))[0]
            name1 = StarName+'_'+ObsDate
            name2 = name[len(name1)+1:]
            title = name1+'\n'+name2
            fig = make_contrast_plot(cm, title, plate_scale=plate_scale, 
                                     contrast_range=(1e-7,5e-2))
            fig.savefig(os.path.join(figure_path, name+'-contrast_map.png'))
            plt.close(fig.number)
    




#!/usr/bin/env python
from astropy.io import fits
import numpy as np
from scipy import ndimage, signal, optimize, linalg, stats
import multiprocessing
import re
import glob
import os
import sys
import time


class Consumer(multiprocessing.Process):

    def __init__(self, task_queue, result_queue):
        multiprocessing.Process.__init__(self)
        self.task_queue = task_queue
        self.result_queue = result_queue

    def run(self):
        proc_name = self.name
        while True:
            next_task = self.task_queue.get()
            if next_task is None:
                # Poison pill means we should exit
                break
            self.result_queue.put(next_task())
        return


class Task(object):
    def __init__(self, index, func, args):
        self.index = index
        self.func = func
        self.args = args

    def __call__(self):
        return self.index, self.func(*self.args)


def _smooth(im, ivar, sig=1, spline_filter=False):
    '''
    Private function _smooth smooths an image accounting for the
    inverse variance.  It optionally spline filters the result to save
    time in later calls to ndimage.map_coordinates.

    Args:
        im: ndarray, 2D image to smooth
        ivar: ndarray, 2D inverse variance, shape should match im
        sig: float, standard deviation of Gaussian smoothing kernel, default 1
        spline_filter: boolean, spline filter the result?  Default False.

    Returns:
        imsmooth : ndarray, smoothed image of the same size as im
    '''

    if not isinstance(im, np.ndarray) or not isinstance(ivar, np.ndarray):
        raise TypeError("image and ivar passed to _smooth must be ndarrays")
    if im.shape != ivar.shape or len(im.shape) != 2:
        raise ValueError("image and ivar must be 2D ndarrays of the same shape")

    nx = int(sig * 4 + 1) * 2 + 1
    x = np.arange(nx) - nx // 2
    x, y = np.meshgrid(x, x)

    window = np.exp(-(x ** 2 + y ** 2) / (2 * sig ** 2))
    imsmooth = signal.convolve2d(im * ivar, window, mode='same')
    imsmooth /= signal.convolve2d(ivar, window, mode='same') + 1e-35
    imsmooth *= im != 0

    if spline_filter:
        imsmooth = ndimage.spline_filter(imsmooth)

    return imsmooth


def _spotloc(phi, sep, pitch=15, D=8.2):
    '''
    Private function _spotloc computes the location of four
    satellite spots in units of lambda

    Args:
        phi: float, the angle of spots in degrees
        sep: float, the separation of the spots in units of lambda/D
        pitch: float, lenslet pitch in units of milliarcseconds. Default 15
        D: float, telescope effective aperture in meters. Default 8.2

    Returns:
        r : ndarray, array of (identical) separations in units of lenslets/microns
        phi : ndarray, array of spot angles in radians
    '''

    phi = np.arange(4) * np.pi / 2 + phi
    r = sep * 1e-6 / D * 3600 * 180 / np.pi / (pitch * 1e-3)
    r = r * np.ones(phi.shape)

    return r, phi


def _par_to_dx_dy(p, lam):
    '''
    Private function _par_to_dx_dy.  Use the parameters given by
    argument p (as a list), together with the wavelength array, to
    compute the offsets as a function of wavelength given as a
    second-order polynomial in wavelength.

    Args:
        p: list of floats, the coefficients of the polynomial fit to the centroid
        lam: wavelength(s), either as a number or as an array

    Returns:
        dx, dy : tuple of 1D arrays, each the same shape as lam, with the wavelength-dependent offsets

    '''

    order = (len(p) - 2) // 2
    dx = (lam * 0 + 1.) * p[0]
    dy = (lam * 0 + 1.) * p[1 + order]

    for i in range(1, order + 1):
        dx += p[i] * lam ** i
        dy += p[1 + i + order] * lam ** i

    return dx, dy


def _spotintens(p, cube, lam):
    '''
    Private function _spotintens computes the negative of the sum of
    the intensity of the four satellite spots induced by the SCExAO DM
    in XYdiag mode.  It is intended to be passed to a minimization
    routine (to maximize sum of the intensities).

    Args:
        p: list of floats
           p[0] is the angle of spots in degrees
           p[1] is the separation in lambda/D
           p[2] - p[-1] are the coefficients of the polynomial fit to the centroid
        cube: 3D ndarray, input data cube, assumed to be smoothed and spline filtered
        lam: 1D ndarray, wavelength array in microns corresponding to the first axis of cube

    Returns:
        sumval : float, the negative of the sum of the spot intensities at the four locations
                 given by p, with integer multiples of pi/2 added to p[0].

    '''

    if not isinstance(cube, np.ndarray):
        raise TypeError("cube must be a 3-dimensional ndarray")
    if len(cube.shape) != 3:
        raise TypeError("cube must be a 3-dimensional ndarray")
    if len(cube) != len(lam):
        raise ValueError("Dimensions of lam and cube must match")

    r, phi = _spotloc(p[0], p[1])
    rcosphi = r * np.cos(phi)
    rsinphi = r * np.sin(phi)
    sumval = 0

    dx, dy = _par_to_dx_dy(p[2:], lam)

    for i in range(cube.shape[0]):
        x = rcosphi * lam[i] + cube.shape[2] // 2 + dx[i]
        y = rsinphi * lam[i] + cube.shape[1] // 2 + dy[i]

        sumval -= np.sum(ndimage.map_coordinates(cube[i], [y, x], prefilter=False))

    return sumval


def _resid(im1, im2):
    '''
    Private function _resid takes two images, and subtracts a constant
    and a scaled version of the second image from the first.  It
    returns the sum of the squared residuals.

    Args:
        im1: ndarray, image
        im2: ndarray, image of the same dimensions as im1

    Returns:
        chisq : float, sum of the squared residuals after subtracting the
                best fit constant+im2 model from im1

    '''

    if not isinstance(im1, np.ndarray) or not isinstance(im2, np.ndarray):
        raise TypeError("Images passed to _resid must be ndarrays.")
    if im1.shape != im2.shape:
        raise ValueError("Dimensions of images in _resid must match.")

    Sxx = np.sum(im2 ** 2)
    Sx = np.sum(im2)
    Sy = np.sum(im1)
    Sxy = np.sum(im1 * im2)
    S = im2.shape[0]
    norm = (S * Sxy - Sx * Sy) / (S * Sxx - Sx ** 2)
    zeropt = (Sxx * Sy - Sx * Sxy) / (S * Sxx - Sx ** 2)

    return np.sum((im1 - norm * im2 - zeropt) ** 2)


def _resid_xy(p, im1, im2, x, y):
    '''
    Private function _resid_xy computes the residual between im1 and
    im2, but with an offset applied to the positions of im2 before
    finding the residual.  The residual is only computed at certain
    points given by ndarrays x and y.  This function is intended to be
    passed to a minimization routine.

    The function calls map_coordinates to compute the reference and
    comparison images at the appropriate points, then calls _resid to
    fit the model and compute the residual.

    Args:
        p: list of two floats
           p[0] is the x offset between im1 and im2
           p[1] is the y offset
        im1: 2D ndarray, reference image, assumed to be spline_filtered
        im2: 2D ndarray, comparison image, assumed to be spline_filtered
        x: 2D ndarray, x coordinates of the points at which to fit im2 to im1
        y: 2D ndarray, y coordinates of the points at which to fit im2 to im1

    Returns:
        chisq : float, square of the residuals between im1 and a model made by
                translating im2 and sampling at x and y.

    '''

    dx, dy = [p[0], p[1]]
    ref_im = ndimage.map_coordinates(im1, [y, x], prefilter=False)
    compare_im = ndimage.map_coordinates(im2, [y + dy, x + dx], prefilter=False)

    return _resid(ref_im, compare_im)


def _cc_resid(p, pp, cube, lam, x, y, retarr=False):
    '''
    Private function _cc_resid cross-correlates wavelength slices of a
    data cube to a reference wavelength.  The output is the sum of the
    squares of the lenslet-by-lenslet differences between the
    reference and scaled/shifted slices (after optimizing the relative
    scale), or the scaled/shifted slices themselves (as a template).

    Args:
        p: list of floats, the coefficients of the polynomial fit to the centroid
        pp: list of floats, the coefficients of the reference polynomial fit to the
            centroid (used to keep the reference center in place when optimizing p)
        cube: 3D ndarray, data cube
        lam: 1D ndarray, wavelengths
        x: ndarray with x lenslet coordinates of the cube to use in the calculation
        y: ndarray with y lenslet coordinates of the cube to use in the calculation
        retarr: return the scaled/shifted template rather than the
                           cross-correlation score? Default False.

    Returns:
        If retarr is False (default), return the sum of the squared differences
        of the various wavelength channels with the reference channel.  If
        retarr is True, return the template image, the average of all of the
        scaled and shifted images at the input x and y coordinates.

    '''

    dx1, dy1 = _par_to_dx_dy(p, lam)
    ny, nx = [cube.shape[1], cube.shape[2]]

    # Centroid of the reference wavelength slice

    iref = cube.shape[0] // 2
    dx2, dy2 = _par_to_dx_dy(pp, lam[iref])
    xx = x + dx2
    yy = y + dy2

    if retarr:
        interparr = np.zeros(tuple([cube.shape[0]] + list(x.shape)))
    else:
        ref = ndimage.map_coordinates(cube[iref], [yy + ny // 2, xx + nx // 2], prefilter=False)
        chisq = 0

    for i in range(cube.shape[0]):

        _x = (xx - dx1[i]) * lam[i] / lam[iref] + dx1[i] + nx // 2
        _y = (yy - dy1[i]) * lam[i] / lam[iref] + dy1[i] + ny // 2

        compare = ndimage.map_coordinates(cube[i], [_y, _x], prefilter=False)
        if retarr:
            interparr[i] = compare
        else:
            chisq += _resid(ref, compare)

    if retarr:
        return np.mean(interparr, axis=0)
    else:
        return chisq


def get_sats_satf(p, cube, lam):
    '''
    retrieves the pixel locations of all four satellite spots at each wavelength,
    and the negative sum of the four spot intensities at each wavelength

    Args:
        p: list of floats, the coefficients of the polynomial fit to the centroid
        cube: ndarray, data cube for which the centroid is fitted
        lam: ndarray, wavelengths for the datacube

    Returns:
        sats: pixel locations (in [x,y] format) of all four satellite spots at each wavelength, shape (wvs, 4, 2)
        satf: float, negative sum of the four spot intensities at each wavelength
    '''

    if not isinstance(cube, np.ndarray):
        raise TypeError("data cube must be a 3-dimensional ndarray")
    if len(cube.shape) != 3:
        raise TypeError("data cube must be a 3-dimensional ndarray")
    if len(cube) != len(lam):
        raise ValueError("Dimensions of lam and cube must match")

    r, phi = _spotloc(p[0], p[1])
    rcosphi = r * np.cos(phi)
    rsinphi = r * np.sin(phi)

    dx, dy = _par_to_dx_dy(p[2:], lam)

    sats = np.zeros((len(lam), 4, 2))
    satf = np.zeros((len(lam), 4))

    for i in range(cube.shape[0]):
        x = rcosphi * lam[i] + cube.shape[2] // 2 + dx[i]
        y = rsinphi * lam[i] + cube.shape[1] // 2 + dy[i]

        # note that for sats, indices are recorded as [x,y], not [y,x], to be consistent with pyklip's convention
        sats[i] = np.stack((x, y), axis=1)
        satf[i] = ndimage.map_coordinates(cube[i], [y, x], prefilter=False)

    return sats, satf


def recen(p, cube, lam, sats, satf, n=None, scale=False, head=None, outfile=None,
          mask=None, data_HDU1=True):
    '''
    Function recen recenters the input data according to the offset
    parameters given in the argument p.  Optionally scale the data by
    wavelength to undo the scaling of diffraction.  Return the
    recentered cube.

    Args:
        p: list of floats, the coefficients of the polynomial fit to the centroid
        cube: 3D ndarray, data cube
        lam: 1D array of wavelengths
        sats: fitted satellite spot indices before recentering
        satf: fitted satellite spot fluxes
        n : integer, spatial dimension of recentered cube.  Default original cube size
        scale: boolean, rescale by wavelength?  Default False
        head: fits header for output file.  Default None
        outfile: string or None, name of output file.  Default None
        mask: boolean lenslet mask
        data_HDU1: boolean, write data to HDU1 and leave HDU0 with no data?  Default True.

    Returns:
        3D ndarray, nlam x n x n, recentered (optionally scaled by wavelength) data cube.

    '''

    if n is None:
        n = cube.shape[1]

    ny, nx = [cube.shape[1], cube.shape[2]]
    dx, dy = _par_to_dx_dy(p[2:], lam)

    x = np.arange(n) - n // 2
    x, y = np.meshgrid(x, x)

    cencube = np.zeros((lam.shape[0], x.shape[0], x.shape[1]))
    for i in range(lam.shape[0]):
        if scale:
            _y = y * lam[i] / lam[len(lam) / 2] + dy[i] + ny // 2
            _x = x * lam[i] / lam[len(lam) / 2] + dx[i] + nx // 2
            print('fitted spots for scaled recen cubes not tested for correctness yet')
            sats[i, :] -= [nx // 2, ny // 2]
            sats[i, :, 1] = sats[i, :, 1] * lam[i] / lam[len(lam) / 2] + ny // 2 - dy[i]
            sats[i, :, 0] = sats[i, :, 0] * lam[i] / lam[len(lam) / 2] + nx // 2 - dx[i]
        else:
            _y = y + dy[i] + ny // 2
            _x = x + dx[i] + nx // 2
            sats[i, :, 1] -= dy[i]
            sats[i, :, 0] -= dx[i]

        cencube[i] = ndimage.map_coordinates(cube[i], [_y, _x], prefilter=False)

    if mask is not None:
        k = (mask.shape[0] - n) // 2
        for i in range(cencube.shape[0]):
            cencube[i] *= mask[k:-k, k:-k]

    if data_HDU1:
        out = fits.HDUList(fits.PrimaryHDU(None, head))
        out.append(fits.PrimaryHDU(cencube))
        out.append(fits.PrimaryHDU(np.ones(cencube.shape)))
    else:
        out = fits.HDUList(fits.PrimaryHDU(cencube, head))

    if outfile is not None:
        out.writeto(outfile, overwrite=True)
        savespot_hdr(sats, satf, outfile, hdu_num=int(data_HDU1))
    else:
        print('output directory not specified, recentered cubes not saved...')

    return cencube


def fitrelcen(image1, image2, x, y, method='Powell'):
    '''
    Function fitrelcen fits for the offset between two images without
    any wavelength dependence by calling _resid_xy, minimizing the sum
    of the squared differences between the images (after optimizing
    over the wavelength-dependent relative normalization).

    Args:
        image1: 2D ndarray, first image
        image2:2D ndarray, second image
        x: ndarray, x coordinates of pixels/lenslets to use
        y: ndarray, y coordinates of pixels/lenslets to use
        method : method passed to scipy.optimize.minimize. Default 'Powell'.

    Returns:
        xc, yc : two floating point numbers giving the best-fit offset between
                 image1 and image2

    '''

    xc, yc = optimize.minimize(_resid_xy, [0, 0],
                               (image1, image2, x, y), method=method).x

    return [xc, yc]


def fitcen(cube, ivar, lam, spotsep=None, guess_center_loc=None, i1=1, i2=-1, r1=15, r2=35, spot_dx=4):
    '''
    Function fitcen.  Fit for the center of a CHARIS data cube using
    the satellite spots by maximizing the agreement between scaled
    cube slices around the spot locations.  If no spot locations are
    provided, use only the diffraction pattern itself in an annulus
    around the image center.

    Args:
        cube: 3D ndarray, CHARIS data cube
        ivar: 3D ndarray, inverse variance of the CHARIS data cube
        lam: 1D ndarray, wavelengths
        spotsep: float or None.  If float, separation of the satellite spots in units
                 of lambda/D.  If None, only use the diffraction pattern in an annulus
                 between r1 and r2.
        guess_center_loc: manually specify initial location of image center if necessary, in [x, y] format
        i1: int, first slice of the data cube to use.  Default 1 (skip slice 0)
        i2: int, high limit of slices to use.  Default -1 (skip last slice)
        r1: float, minimum separation from approximate center for the annulus of the
            diffraction pattern to use in centroiding.  Default 15
        r2: float, maximum separation from approximate center for the annulus of the
            diffraction pattern to use in centroiding.  Default 35
        spot_dx: float, radius around spot location to cut out in order to match the
                 spot location as a function of wavelength.  Default 4

    Returns:
        p : list of floats
            p[0] is the angle of spots in degrees
            p[1] is the separation in lambda/D
            p[2] - p[-1] are the coefficients of the polynomial fit to the centroid

    '''

    ####################################################################
    # Lightly smooth the cube before starting.
    ####################################################################

    cubesmooth = cube.copy()
    mask = np.any(cubesmooth, axis=0) != 0
    iref = i1 + len(lam[i1:i2]) // 2

    for i in range(cube.shape[0]):
        cubesmooth[i] = _smooth(cube[i], ivar[i], lam[i] / 3., spline_filter=False)
        cubesmooth[i] *= mask
        cubesmooth[i] = ndimage.spline_filter(cubesmooth[i])
    # print(np.mean(cube), np.mean(cubesmooth), np.std(cube), np.std(cubesmooth))

    if spotsep is not None:
        if np.abs(spotsep - 15.9) < 1:
            phi = -18 * np.pi / 180
        elif np.abs(spotsep - 10) < 1:
            phi = 27 * np.pi / 180
        else:
            print("Must call fitcen with a valid separation for the satellite spots.")
            print("Currently accepted separations: 10 and 15.5 lambda/D.")
            return None

    ####################################################################
    # Initial center using inner region of the cube (not the spots)
    ####################################################################


    x = np.arange(cube.shape[1]) - cube.shape[1] // 2
    x, y = np.meshgrid(x, x)
    indx = np.where((x ** 2 + y ** 2 > r1 ** 2) * (x ** 2 + y ** 2 < r2 ** 2))

    if guess_center_loc is None:
        xc0, yc0 = optimize.minimize(_cc_resid, [0, 0], ([0, 0], cubesmooth[i1:i2], lam[i1:i2], x[indx], y[indx], False),
                                     method='Powell').x
    else:
        xc0 = guess_center_loc[0]
        yc0 = guess_center_loc[1]

    if spotsep is None:
        return [xc0, yc0]

    p0 = [phi, spotsep, xc0, 0, yc0, 0]

    ####################################################################
    # Now estimate the center by fitting for the spot locations at
    # each wavelength
    ####################################################################

    p1 = optimize.minimize(_spotintens, p0, (cubesmooth[i1:i2], lam[i1:i2]), method='Powell').x
    phi, spotsep = [p1[0], p1[1]]

    ####################################################################
    # Use the areas immediately around the spots to get a center, now
    # by cross-correlation.
    ####################################################################

    _r, _phi = _spotloc(phi, spotsep)
    rcosphi = _r * np.cos(_phi) * lam[iref]
    rsinphi = _r * np.sin(_phi) * lam[iref]

    ok = np.zeros(x.shape)
    for i in range(len(_phi)):
        ok += (rcosphi[i] - x) ** 2 + (rsinphi[i] - y) ** 2 < spot_dx ** 2
    indx = np.where(ok != 0)

    p2 = optimize.minimize(_cc_resid, p0[2:], (p1[2:], cubesmooth[i1:i2], lam[i1:i2], x[indx], y[indx]),
                           method='Powell').x

    ####################################################################
    # Pull out the region of the image corresponding to the average spots.
    ####################################################################

    avgspots = _cc_resid(p2, p2, cubesmooth[i1:i2], lam[i1:i2], x, y, retarr=True)

    lam0 = lam[iref]

    ####################################################################
    # Revise the center location one more time, use this new center
    # (derived from the spots) to shift the coefficients of the fit
    ####################################################################

    p3 = optimize.minimize(_spotintens, [phi, spotsep, 0, 0], (np.asarray([avgspots]), np.asarray([lam0])),
                           method='Powell').x

    ####################################################################
    # Use the center derived from the spots to shift the coefficients
    # of the fit
    ####################################################################

    # center:

    # x: (p2[0] + p3[2] + p2[1]*lam0 - p2[0] - p2[1]*lam)*lam/lam0 + p2[0] + p2[1]*lam
    # y: (p2[2] + p3[3] + p2[3]*lam0 - p2[2] - p2[3]*lam)*lam/lam0 + p2[2] + p2[3]*lam

    # x: p2[0] + (p3[2]/lam0 + 2*p2[1])*lam - (p2[1]/lam0)*lam**2
    # y: p2[2] + (p3[3]/lam0 + 2*p2[3])*lam - (p2[3]/lam0)*lam**2

    phi, spotsep = [p3[0], p3[1]]
    cen_coef = [p2[0], p3[2] / lam0 + 2 * p2[1], -p2[1] / lam0,
                p2[2], p3[3] / lam0 + 2 * p2[3], -p2[3] / lam0]

    return list([phi, spotsep]) + cen_coef


def fitcen_parallel(infiles, astrogrid_status=None, astrogrid_sep=None, smooth_coef=True,
                    guess_center_loc=None, maxcpus=multiprocessing.cpu_count() // 2):
    '''
    Function fitcen_parallel.  Centroid a series of CHARIS data cubes
    in parallel using fitcen.  By default, get the wavelengths and
    astrogrid parameters from the headers.  This might fail on early
    CHARIS data before the standardization of headers.

    Args:
        infiles: list of file names for CHARIS data cubes
        astrogrid_status: None or list of astrogrid configurations for SCExAO.
                          If None, try to read the astrogrid configuration from the header.
                          If this fails, assume there is no astrogrid and centroid using the
                          general diffraction pattern.  Default None.
        astrogrid_sep: None or list of astrogrid spot separations in units of lambda/D.
                       If None, try to read from the header.
                       If that fails, centroid using the general diffraction pattern.
                       Default None.
        smooth_coef: boolean.  smooth the nonlinear coefficients of the centroid fit (the terms
                     proportional to lambda and lambda^2) over the sequence of cubes?
                     Default True.
        guess_center_loc: manually specify initial location of image center if necessary, in [x, y] format
        maxcpus: int, maximum number of CPUs to use in parallelization

    Returns:
        [centroid_params, x, y, mask]

        centroid_params : 2D array of centroid parameters, first dimension is the number of files.
                          Second dimension is the length of theh wavelength-dependent model.
        x : x-coordinates of the centroid at the middle wavelength
        y : y-coordinates of the centroid at the middle wavelength
        mask : 1D boolean array, True if astrogrid was on, or None if the astrogrid was never on.

    '''

    ####################################################################
    # First try to load the astrogrid status and spot separations from
    # the FITS headers
    ####################################################################

    if astrogrid_status is None:
        astrogrid_status = []
        astrogrid_sep = []
        for ifile in infiles:
            head = fits.open(ifile)[0].header
            try:
                astrogrid_status += [head['X_GRDST']]
                astrogrid_sep += [head['X_GRDSEP']]
            except:
                astrogrid_status += [None]
                astrogrid_sep += [None]

    tasks = multiprocessing.Queue()
    results = multiprocessing.Queue()
    ncpus = min(multiprocessing.cpu_count(), maxcpus)
    consumers = [Consumer(tasks, results)
                 for i in range(ncpus)]
    for w in consumers:
        w.start()

    fids = [int(re.sub('.*CRSA', '', re.sub('_cube.fits', '', infile)))
            for infile in infiles]

    lamlist = []
    grid_on = np.zeros(len(fids), np.int32)

    ####################################################################
    # Load each file, get the wavelengths, and pass it to fitcen.
    ####################################################################

    for i in range(len(fids)):

        hdulist = fits.open(infiles[i])
        cube = hdulist[1].data
        ivar = hdulist[2].data
        head = hdulist[0].header
        lam = head['lam_min'] * np.exp(np.arange(cube.shape[0]) * head['dloglam'])
        lam *= 1e-3
        lamlist += [lam]

        if astrogrid_status[i] == 'XYdiag':
            spotsep = astrogrid_sep[i]
            grid_on[i] = 1
        elif astrogrid_status[i] is None:
            spotsep = 15.5
        else:
            spotsep = None

        tasks.put(Task(i, fitcen, (cube, ivar, lam, spotsep, guess_center_loc, 1, -1)))

    for i in range(ncpus):
        tasks.put(None)

    centroid_params = None

    for fid in fids:
        index, result = results.get()
        if centroid_params is None:
            centroid_params = np.zeros((len(fids), len(result)))
        centroid_params[index] = result

    x = None
    y = None

    if np.sum(grid_on) == 0:
        mask = None
    else:
        mask = grid_on

    ####################################################################
    # If desired, smooth the nonlinear coefficients (in wavelength) of
    # the centroiding polynomial so that it varies smoothly with time.
    # Do not smooth the offsets, i.e., do not change the centers at a
    # given reference wavelength.  These can be refined later.
    ####################################################################

    if smooth_coef:

        x1 = polyfit(fids, centroid_params[:, 3], mask=mask, return_y=False)
        x2 = polyfit(fids, centroid_params[:, 4], mask=mask, return_y=False)
        y1 = polyfit(fids, centroid_params[:, 6], mask=mask, return_y=False)
        y2 = polyfit(fids, centroid_params[:, 7], mask=mask, return_y=False)
        x = np.zeros(centroid_params.shape[0])
        y = np.zeros(centroid_params.shape[0])

        for i in range(len(fids)):
            p = centroid_params[i]
            lamref = lamlist[i][len(lamlist[i]) // 2]
            dxref = p[3] * lamref + p[4] * lamref ** 2
            dyref = p[6] * lamref + p[7] * lamref ** 2
            dxnew = x1[i] * lamref + x2[i] * lamref ** 2
            dynew = y1[i] * lamref + y2[i] * lamref ** 2

            x[i] = dxref + p[2]
            y[i] = dyref + p[5]

            centroid_params[i, 2] += dxref - dxnew
            centroid_params[i, 5] += dyref - dynew

        centroid_params[:, 3] = x1
        centroid_params[:, 4] = x2
        centroid_params[:, 6] = y1
        centroid_params[:, 7] = y2

    return [centroid_params, x, y, mask]


def fitallrelcen(infiles, r1=15, r2=50, maxcpus=multiprocessing.cpu_count() // 2):
    '''
    Function fitallrelcen.  Fit for the relative centroids between all
    pairs of frames at the central wavelength using the PSF in an
    annulus around the center.  Return the best-fit relative offets.

    Args:
        infiles: list of file names
        r1: int, minimum separation in lenslets from the image center for annular reference region.  Default 15.
        r2: int, maximum separation in lenslets from the image center for annular reference region.  Default 50.
        maxcpus: int, maximum number of CPUs to allocate for parallelization. Default 1/2 of the available CPUs.

    Returns:
        xsol : 1D ndarray of the relative centers in x
        ysol : 1D ndarray of the relative centers in y

    '''

    ncpus = min(multiprocessing.cpu_count(), maxcpus)

    shape = fits.open(infiles[0])[1].data.shape
    iref = shape[0] // 2

    ####################################################################
    # Lightly smooth all images first.
    ####################################################################

    allims = np.zeros((len(infiles), shape[1], shape[2]))

    for i in range(len(infiles)):
        im = fits.open(infiles[i])[1].data[iref]
        ivar = fits.open(infiles[i])[2].data[iref]
        allims[i] = _smooth(im, ivar, 0.5, True)

    tasks = multiprocessing.Queue()
    results = multiprocessing.Queue()
    consumers = [Consumer(tasks, results)
                 for k in range(ncpus)]
    for w in consumers:
        w.start()

    ####################################################################
    # Find the relative centers by cross-correlation, using only
    # lenslets >r1 and <r2 from the nominal center.
    ####################################################################

    ny, nx = [allims.shape[1], allims.shape[2]]
    x = np.arange(nx) - nx // 2
    x, y = np.meshgrid(x, x)
    r = np.sqrt(x ** 2 + y ** 2)
    x += nx // 2
    y += ny // 2

    ok = np.where((r > r1) * (r < r2))
    y = y[ok].copy()
    x = x[ok].copy()

    for i in range(len(infiles)):
        for j in range(i + 1, len(infiles)):
            index = i * (len(infiles)) + j
            tasks.put(Task(index, fitrelcen, (allims[i], allims[j], x, y)))

    for k in range(ncpus):
        tasks.put(None)

    ####################################################################
    # Once we have all offsets between pairs of frames, solve for the
    # best-fit positions.  This is only defined up to a constant, so
    # also add the constraint that the mean position is zero.  We can
    # add an offset later based on the absolute centroiding routine.
    ####################################################################

    arr_x = np.zeros((len(infiles), len(infiles)))
    arr_y = np.zeros((len(infiles), len(infiles)))

    for i in range(len(infiles)):
        for j in range(i + 1, len(infiles)):
            index, result = results.get()
            fid2 = index % len(infiles)
            fid1 = index // len(infiles)

            arr_x[fid1, fid2] = result[0]
            arr_x[fid2, fid1] = -result[0]
            arr_y[fid1, fid2] = result[1]
            arr_y[fid2, fid1] = -result[1]

    n = len(infiles)
    b = np.zeros((n + 1))
    b[:n] = np.sum(arr_x, axis=0)
    arr = np.ones((n, n + 1))
    arr[:, :n] *= -1
    arr[:, :n] += n * np.identity(n)
    xsol = np.linalg.lstsq(arr.T, b, rcond=-1)[0]

    b[:n] = np.sum(arr_y, axis=0)
    ysol = np.linalg.lstsq(arr.T, b, rcond=-1)[0]

    return xsol, ysol


def polyfit(x, y, order=2, clip=2.5, niter=5, mask=None, return_y=True):
    '''
    Smooth a series of points with a polynomial, iteratively clipping
    outliers.

    Args:
        x: 1D ndarray of x coordinates for the polynomial fit
        y: 1D ndarray of y coordinates for the polynomial fit
        order: int, order of the polynomial to fit.  Default 2.
        clip: float, number of sigma outliers to clip.  Default 2.5.
        niter: int, number of iterations of sigma clipping.  Default 5.
        mask: boolean ndarray or None: mask each y value?  Default None
        return_y: boolean, return smoothed y values (as opposed to coefficients of the polynomial fit)?
                  Default True.

    Returns:
        y_smoothed : 1D array, if return_y=True
        coef : array of the polynomial coefficients if return_y=False

    '''

    if len(x) <= order and not return_y:
        return y

    arr = np.ones((len(x), order + 1))
    for i in range(1, order + 1):
        arr[:, i] = (x - np.median(x)) ** i

    ok = np.ones(y.shape)
    fit_vals = y.copy()

    for i in range(niter):

        _arr = arr.copy()
        if mask is not None:
            ok *= mask
        for j in range(order + 1):
            _arr[:, j] *= ok

        fitcoef = np.linalg.lstsq(_arr, y * ok, rcond=-1)[0]

        resid = y.copy()
        fit_vals[:] = 0
        for j in range(order + 1):
            resid -= fitcoef[j] * arr[:, j]
            fit_vals += fitcoef[j] * arr[:, j]

        std = np.sqrt(np.mean(resid[np.where(ok)] ** 2))
        ok = resid ** 2 < clip ** 2 * std ** 2

    if return_y:
        return fitcoef
    else:
        return fit_vals


def specphotcal(infiles, cencoef, aperture=1.):
    '''
    Function specphotcal.  Computes approximate photometry from the
    satellite spots (using aperture photometry) and scale each
    wavelength to this photometric value.  This should crudely put the
    cubes in units of contrast, though it omits the scaling of
    satellite spot intensity with 1/lambda^2.

    Args:
        infiles: list of file names with CHARIS data cubes
        cencoef: 2D ndarray with coefficeitns
        aperture: float, radius of aperture for photometry in units of lambda/D

    Returns:
        all_phot: photocalibration coefficients

    '''

    fids = [int(re.sub('.*CRSA', '', re.sub('_cube.fits', '', infile)))
            for infile in infiles]

    phi = polyfit(fids, cencoef[:, 0], return_y=False)
    sep = polyfit(fids, cencoef[:, 1], return_y=False)

    all_phot = []
    all_x = []
    all_y = []

    for i in range(len(fids)):

        hdulist = fits.open(infiles[i], mode='update')
        im = hdulist[1].data
        head = hdulist[0].header
        lam = head['lam_min'] * np.exp(np.arange(im.shape[0]) * head['dloglam'])
        lam *= 1e-3

        ny, nx = [im.shape[1], im.shape[2]]

        dx, dy = _par_to_dx_dy(cencoef[i, 2:], lam)
        phot = np.zeros(lam.shape)

        x = np.arange(nx)
        y = np.arange(ny)
        x, y = np.meshgrid(x, y)
        _r, _phi = _spotloc(phi[i], sep[i])

        # Manual offsets to account for the fact that two of the spots
        # lie atop spiders.  Check to see if the first spot is atop a
        # spider or not.

        if np.abs(_phi[0] % np.pi - 2.8) < np.pi / 4:
            _dphi = -np.pi / 2 + 12 * np.pi / 180 * np.asarray([1, -1, 1, -1])
        else:
            _dphi = -np.pi / 2 + 12 * np.pi / 180 * np.asarray([-1, 1, -1, 1])

        caltable = []

        for j in range(lam.shape[0]):

            xcoord = np.cos(_phi) * _r * lam[j] + dx[j] + nx // 2
            ycoord = np.sin(_phi) * _r * lam[j] + dy[j] + ny // 2
            xref = np.cos(_phi + _dphi) * _r * lam[j] + dx[j] + nx // 2
            yref = np.sin(_phi + _dphi) * _r * lam[j] + dy[j] + ny // 2

            # CHARIS is approximately critically sampled at 1.15 microns
            radius = 2 * aperture * (lam[j] / 1.15)


            for k in range(len(xcoord)):
                dist = np.sqrt((x - xcoord[k]) ** 2 + (y - ycoord[k]) ** 2)
                distref = np.sqrt((x - xref[k]) ** 2 + (y - yref[k]) ** 2)

                ref = np.sum(im[j][distref < radius])
                phot[j] += np.sum(im[j][dist < radius]) - ref

            caltable += [(dx[j] + nx // 2, dy[j] + ny // 2, phot[j])]
        all_phot += [phot]
        all_x += [dx + nx // 2]
        all_y += [dy + ny // 2]

    return all_phot


import numpy as np
import os
from sys import stdout
from astropy.io import fits
import pyklip.klip as klip

class PSFLibrary(object):
    """
    This is an PSF Library to use for reference differential imaging

    Attributes:
        master_library (np.ndarray): aligned library of PSFs. 3-D cube of dim = [N, y, x]. Where N is ALL files
        aligned_center (array-like): a (x,y) coordinate specifying common center the library is aligned to
        master_filenames (np.ndarray): array of N filenames for each frame in the library. Should match with
                                pyklip.instruments.Data.filenames for cross-matching
        master_correlation (np.ndarray): N x N array of correlations between each 2 frames
        master_wvs (np.ndarray): N wavelengths for each frame
        nfiles (int): the number of files in the PSF library
        dataset (pyklip.instruments.Instrument.Data)
        correlation (np.ndarray): N_data x M array of correlations between each 2 frames where M are the selected frames
                            and N_data is the number of files in the dataset. Along the N_data dimension, files are
                            ordered in the same way as the dataset object
        isgoodpsf (np.ndarray): array of N indicating which M PSFs are good for this dataset

    """

    def __init__(self, data, aligned_center, filenames, correlation_matrix=None, wvs=None, compute_correlation=False):
        """

        Args:
            data (np.ndarray): a 3-D cube of PSF library files (dim = [N, y, x]) where N is number of files
                   These files should have already been registered to a common center
            aligned_center (array-like): an (x,y) coordinate specifying the common center all files are registered to
            filenames (np.ndarray): a array of N filenames for each file. These should be in the same format as a
                        pyklip.instruments.Instrument.Data.filenames array so that the two can be cross-matched
            correlation_matrix (np.ndarray): an N x N matrix that expresses the correlation between each two frames in library
            wvs (np.ndarray): array of N wavelengths that correspond to the wavelengths of the library
            compute_correlation (boolean): if True, compute the correlation matrix. Note that this can potentially take a
                                    long time, so you really should be doing it once and saving it
        """
        # call init() of super class
        super(PSFLibrary, self).__init__()

        # Do some checking to make sure the data, filenames list and correlation_matrix (if it exists)
        # all have the same number of files. 
        nfiles_data = np.shape(data)[0]
        nfiles_list = np.shape(filenames)[0]

        if nfiles_data != nfiles_list:
            raise AttributeError("The number of files in the data array and filenames list aren't the same. Something is wrong")

        if correlation_matrix is not None:
            nfiles_correlation = np.shape(correlation_matrix)[0]
            if nfiles_correlation != nfiles_data: 
                raise AttributeError("The number of files in the correlation matrix and in the data array aren't the same. Something is wrong")

        # generate master list of files and meta data from inputs
        self.master_library = data
        self.aligned_center = aligned_center
        self.master_filenames = np.asarray(filenames)

        self.master_correlation = correlation_matrix
        self.correlation_mask = None
        self.master_wvs = wvs
        self.nfiles = nfiles_data
        # fields in the context of a specific dataset
        self.dataset = None
        self.correlation = None
        self.isgoodpsf = None

        # check if correlation matrix was passed in
        if correlation_matrix is None and not compute_correlation:
            raise AttributeError("You didn't pass in a correlation matrix, which means it needs to be computed. Are you "
                                 "sure you want to do this? (This may take a while if you have 10,000+ files)")
        elif compute_correlation:
            self._compute_correlation()

    def _compute_correlation(self, verbose=False, force=False, mask=None):
        """
        Computes the correlation matrix and saves it in self.master_correlation

        This computes the correlation between two psfs for every pair of psfs in the library. 
        This correlation matrix can be used for the selection of PSFs ahead of time, but as of right now
        it isn't super useful for saving time (e.g. when there are subsections the covariance still needs
        to be recomputed with all the PSFs currently being used).

        The mask should have 0s where you want to correlate and NANs when you don't. 

        """

        #Get the number of files

        if np.size(self.master_correlation) > 1: 
            print("WARNING: your mater_correlation matrix already has data in it")

            if not force:
                print("WARNING: If you want to overwrite the correlation matrix set the 'force' flag to True")
                return
            else: 
                print("WARNING: overwriting master_correlation")


        self.master_correlation=np.zeros([self.nfiles,self.nfiles])

        if verbose:
            print("Making correlation matrix")

        if mask != None:
            self.correlation_mask = mask

        #Loop the correlation matrix calculation
        for i in np.arange(0, self.nfiles):
            self.master_correlation[i,i]=1.

            #TODO: PARALLELIZE THIS STEP. 

            #Cycle through every file that comes AFTER the current file 
            for j in np.arange(i+1, self.nfiles):

                if verbose:
                    # print("Correlating file "+ str(i) + " with file "+str(j) + "  \r")
                    stdout.write("\r Correlating file {0} with file {1}".format(i,j))
                    stdout.flush()
                
                #You might want to only correlate some of the image. 
                if mask != None:
                    where_to_corr = (self.master_library[i,:,:] == self.master_library[i,:,:]) & (self.master_library[j,:,:] == self.master_library[j,:,:]) & (mask == mask)
                else: 
                #Ditch where either of the two arrays have NANs
                    where_to_corr = (self.master_library[i,:,:] == self.master_library[i,:,:]) & (self.master_library[j,:,:] == self.master_library[j,:,:]) 

                data1= self.master_library[i,:,:]
                data2= self.master_library[j,:,:]

                #I believe this bit was copied and pasted from pyklip at some point. 
                covar_psfs=np.cov([data2[where_to_corr], data1[where_to_corr]])
                covar_diag = np.diagflat(1./np.sqrt(np.diag(covar_psfs)))
                corr_psfs = np.dot( np.dot(covar_diag, covar_psfs ), covar_diag)

                self.master_correlation[i,j]=corr_psfs[0,1]
                self.master_correlation[j,i]=corr_psfs[0,1]

        if verbose:
            print("Done making correlation matrix")


    def save_correlation(self, filename, overwrite=False, clobber=None, format="fits"):
        """
        Saves self.correlation to a file specified by filename
        Args:
            filename (str): filepath to store the correlation matrix
            overwrite (bool): if true overwrite the previous correlation matrix
            clobber (bool): same as overwrite, but deprecated in astropy. 
            format (str): type of file to store the correlation matrix as. Supports numpy?/fits?/pickle? (TBD)

        """
        
        if clobber is not None:
            print("pyklip.rdi.save_correlation: clobber was deprecated in Astropy version 2.0 and" \
            "will be removed in a future version. Use argument overwrite instead)")
            overwrite = clobber


        #TODO: We should probably save more information into the header here, but what exactly it'll be is TBD
        if format == "fits":
            hdu = fits.PrimaryHDU(self.master_correlation)
            if os.path.isfile(filename):
            #If the file already exists give user warning.     
                if overwrite:
                    hdu.writeto(filename, overwrite=overwrite)
                else: 
                    print("save_correlation: File already exists. Set overwrite=True to overwrite")
            else:
                hdu.writeto(filename)

        #But for now only fits
        else:
            print("Sorry, fits is the only filetype type currently supported for saving correlation matrices")
            
    
    def prepare_library(self, dataset, badfiles=None):
        """
        Prepare the PSF Library for an RDI reduction of a specific dataset by only taking the part of the
        library we need.

        Args:
            dataset (pyklip.instruments.Instrument.Data):
            badfiles (np.ndarray): a list of filenames corresponding to bad files we want to also exclude

        Returns:

        """

        # we need to exclude bad files and files already in the dataset itself (since that'd be ADI/SDI/etc)
        # strip away the directories in the master_filenames
        master_just_filenames = np.asarray([filename.split(os.sep)[-1] for filename in self.master_filenames])
        dataset_just_filenames = np.asarray([filename.split(os.sep)[-1] for filename in dataset.filenames])
        # print(dataset_just_filenames)
        # compare with the dataset filnames (also d)
        in_dataset = np.in1d(master_just_filenames, dataset_just_filenames)
        
        # don't compare directly with None
        if badfiles is None:
            badfiles = np.full(np.shape(self.master_filenames), False, dtype=bool)
        are_bad = np.in1d(self.master_filenames, badfiles)
        
        # good ones are the ones that don't fall in either category
        isgood = ~in_dataset & ~badfiles
        good = np.where(isgood)[0]

        # create a view on the good files

        # figure out how the ordering of dataset files are in the PSF library compared to the dataset
        # we want to match the dataset
        # filenames_of_dataset_in_lib = self.master_filenames[np.where(in_dataset)]
        filenames_of_dataset_in_lib = self.master_filenames[in_dataset]
        dataset_file_indices_in_lib = []
        for filename in filenames_of_dataset_in_lib:
            index = np.where(filename == self.master_filenames)[0][0]
            dataset_file_indices_in_lib.append(index)

        if np.size(dataset_file_indices_in_lib) < 1:
            print("Dataset not found in PSF Library, library not prepared.")
        else:
            dataset_file_indices_in_lib = np.array(dataset_file_indices_in_lib)
            # generate a correlation matrix that's N_dataset x N_goodpsfs
            # the ordering of the correlation matrix also ensures that N_dataset is ordered the same as datasets
            self.correlation = self.master_correlation[dataset_file_indices_in_lib]

            # generate a list indicating which files are good
            self.isgoodpsf = good

    def add_new_dataset_to_library(self, dataset, collapse = False, verbose=False):
        """
        Add all the files from a new dataset to the PSF library and add them to the correlation matrix. 
        If a mask was used for the correlation matrix, use it here too. 

        NOTE: This routine already assumes that the data has been centered. 

        Args:
            dataset (pyklip.instruments.Instrument.Data)
        """

        if collapse: 
            #Collapse the dataset
            if verbose:
                stdout.write("Collapsing spectral cubes.....")
                stdout.flush()
            dataset.spectral_collapse(align_frames=False)
            if verbose:
                stdout.write("\rCollapsing spectral cubes.....Done\n")

        n_newfiles = dataset.input.shape[0]
        if verbose:
            print("Found {} new files".format(n_newfiles))

        if verbose:
            stdout.write("Appending to master_library and master_filenames arrays.....")
            stdout.flush()

        #Increase the size of the data array, correlation matrix and file list
        self.master_correlation = np.pad(self.master_correlation,((0,n_newfiles),(0,n_newfiles)), mode='constant', constant_values=0)
        self.master_library = np.pad(self.master_library,((0,n_newfiles),(0,0),(0,0)), mode='constant', constant_values=0)

        #Add the filenames to the library
        self.master_filenames = np.append(self.master_filenames,dataset.filenames)        

        #Put the new data in the master_library
        self.master_library[self.nfiles:self.nfiles+n_newfiles] = dataset.input

        if verbose:
            stdout.write("Appending to master_library and master_filenames arrays.....Done\n")
            stdout.flush()

        if verbose:
            print("Correlating {} new files with existing {} files in the library".format(n_newfiles,self.nfiles))

        #Run the correlation
        for i in np.arange(self.nfiles+n_newfiles-1,self.nfiles-1,-1):
            self.master_correlation[i,i]=1.

            #TODO: PARALLELIZE THIS STEP

            #Cycle through every file that comes AFTER the current file 
            for j in np.arange(0,i):

                if verbose:
                    # print("Correlating file "+ str(i) + " with file "+str(j) + "  \r")
                    stdout.write("\r Correlating new file {0} with file {1}".format(n_newfiles - (i-self.nfiles+1),j))
                    stdout.flush()
                
                #You might want to only correlate some of the image. 
                if self.correlation_mask != None:
                    where_to_corr = (self.master_library[i,:,:] == self.master_library[i,:,:]) & (self.master_library[j,:,:] == self.master_library[j,:,:]) & (self.correlation_mask == self.correlation_mask)
                else: 
                #Ditch where either of the two arrays have NANs
                    where_to_corr = (self.master_library[i,:,:] == self.master_library[i,:,:]) & (self.master_library[j,:,:] == self.master_library[j,:,:]) 

                data1= self.master_library[i,:,:]
                data2= self.master_library[j,:,:]

                #I believe this bit was copied and pasted from pyklip at some point. 
                covar_psfs=np.cov([data2[where_to_corr], data1[where_to_corr]])
                covar_diag = np.diagflat(1./np.sqrt(np.diag(covar_psfs)))
                corr_psfs = np.dot( np.dot(covar_diag, covar_psfs ), covar_diag)

                self.master_correlation[i,j]=corr_psfs[0,1]
                self.master_correlation[j,i]=corr_psfs[0,1]
        self.nfiles = self.nfiles+n_newfiles
        if verbose:
            print("\nDone updating correlation matrix")




import numpy as np



def _matern32(x, y, sigmas, corr_len):
    """
    Generates a Matern (nu=3/2) covariance matrix that assumes x/y has the same correlation length

    C_ij = \sigma_i \sigma_j (1 + sqrt(3) r_ij / l) exp(-sqrt(3) r_ij / l)

    Args:
        x (np.array): 1-D array of x coordinates
        y (np.array): 1-D array of y coordinates
        sigmas (np.array): 1-D array of errors on each pixel
        corr_len (float): correlation length of the Matern function

    Returns:
        cov (np.array): 2-D covariance matrix parameterized by the Matern function
    """
    r = np.sqrt((x[:, None] - x[None, :])**2 + (y[:, None] - y[None, :])**2)
    arg = np.sqrt(3) * r / corr_len
    cov = sigmas[:, None] * sigmas[None, :] * (1+arg) * np.exp(-arg)
    return cov


def matern32(x, y, sigmas, corr_len):
    """
    Generates a Matern (nu=3/2) covariance matrix that assumes x/y has the same correlation length

    C_ij = \sigma_i \sigma_j (1 + sqrt(3) r_ij / l) exp(-sqrt(3) r_ij / l)

    Args:
        x (np.array): 1-D array of x coordinates
        y (np.array): 1-D array of y coordinates
        sigmas (np.array): 1-D array of errors on each pixel
        corr_len (float): correlation length of the Matern function

    Returns:
        cov (np.array): 2-D covariance matrix parameterized by the Matern function
    """
    return _matern32(x, y, sigmas, corr_len)


def _sq_exp(x, y, sigmas, corr_len):
    """
    Generates square exponential covariance matrix that assumes x/y has the same correlation length

    C_ij = \sigma_i \sigma_j exp(-r_ij^2/[2 l^2])

    Args:
        x (np.array): 1-D array of x coordinates
        y (np.array): 1-D array of y coordinates
        sigmas (np.array): 1-D array of errors on each pixel
        corr_len (float): correlation length (i.e. standard deviation of Gaussian)

    Returns:
        cov (np.array): 2-D covariance matrix parameterized by the Matern function
    """
    r = np.sqrt((x[:, None] - x[None, :])**2 + (y[:, None] - y[None, :])**2)
    arg = r**2 / (2 * corr_len**2)
    cov = sigmas[:, None] * sigmas[None, :] * np.exp(-arg)
    return cov


def sq_exp(x, y, sigmas, corr_len):
    """
    Generates square exponential covariance matrix that assumes x/y has the same correlation length

    C_ij = \sigma_i \sigma_j exp(-r_ij^2/[2 l^2])

    Args:
        x (np.array): 1-D array of x coordinates
        y (np.array): 1-D array of y coordinates
        sigmas (np.array): 1-D array of errors on each pixel
        corr_len (float): correlation length (i.e. standard deviation of Gaussian)
        mode (string): either "numpy", "cython", or None, specifying the implementation of the kernel.

    Returns:
        cov (np.array): 2-D covariance matrix parameterized by the Matern function
    """
    return _sq_exp(x, y, sigmas, corr_len)

def delta(x, y, sigmas, *args):
    """
    Generates a diagonal covariance matrix

    C_ij = \sigma_i \sigma_j delta_{ij}

    Args:
        x (np.array): 1-D array of x coordinates
        y (np.array): 1-D array of y coordinates
        sigmas (np.array): 1-D array of errors on each pixel
    """
    return np.diag(sigmas**2)


import numpy as np
from astropy import wcs

def generate_wcs(parangs, center, flipx=False, platescale=0.010, radec=None):
    """
    Generates a "good-enough" astropy.wcs object based on the parallactic angle and the center

    Args:
        parangs (float): angle to rotate the image by to align North up (degrees). Typically, this is dataset.PAs
        center (tuple): [x,y] pixel location for the star center
        flipx (bool): if True, the image is in a right-handed (North-up, East-right) coordinate system. Default is False.
        platescale (float): platescale of image in arcsec/pixel. This is just cosmetic, as pyKLIP only uses wcs for rotation.
        radec (tuple): RA and Dec coordinates of the object (degrees)

    Returns:
        astropy.wcs.WCS object with the orientation of each image
    """
    vert_ang = np.radians(parangs)
    if flipx:
        negsign = -1
    else:
        negsign = 1
    pc = np.array([[-negsign * np.cos(vert_ang), negsign * np.sin(vert_ang)],[np.sin(vert_ang), np.cos(vert_ang)]])
    cdmatrix = pc * platescale /3600.

    new_hdr = {}
    new_hdr['CD1_1'] = cdmatrix[0,0]
    new_hdr['CD1_2'] = cdmatrix[0,1]
    new_hdr['CD2_1'] = cdmatrix[1,0]
    new_hdr['CD2_2'] = cdmatrix[1,1]
    
    new_hdr['CRPIX1'] = center[0]
    new_hdr['CRPIX2'] = center[1]

    new_hdr['CTYPE1'] = 'RA---TAN'
    new_hdr['CTYPE2'] = 'DEC--TAN'

    new_hdr['CDELT1'] = platescale / 3600
    new_hdr['CDELT2'] = platescale / 3600

    if radec is not None:
        new_hdr['CRVAL1'] = radec[0]
        new_hdr['CRVAL2'] = radec[1]
    else:
        new_hdr['CRVAL1'] = 0 
        new_hdr['CRVAL2'] = 0

    w = wcs.WCS(new_hdr)

    return w
    


import numpy as np
import scipy
import pyklip.klip

def _img_xcorr_shift(shift, frame1, frame2, mask=None):
    """
    Compute the negative of the xcorr between frame1 and frame2 given shift:
    
    Args:
        shift: [dx, dy] in pixels to shift frame2
        frame1: refernce frame of shape (Y, X)
        frame2: frame to align of shape (Y, X)
        mask: shape of (Y, X) where =1 for good pixels and =0 for bad pixels
        
    Return:
        negative of xcorr between frame1 and the shifted frame2
    """
    dx, dy = shift
    oldcenter = [frame2.shape[1], frame2.shape[0]]
    frame2_shifted = pyklip.klip.align_and_scale(frame2, [oldcenter[0] + dx, oldcenter[1] + dy], oldcenter)
    if mask is not None:
        good = np.where(mask > 0)
        frame1 = frame1[good]
        frame2_shifted = frame2_shifted[good]
    
    return -np.nansum(frame1*frame2_shifted)


def find_best_shift(frame, ref_frame, guess_center=None, inner_mask=5, outer_mask=None):
    """
    Finds the best shift based on xcorr
    
    Args:
        frame: frame to find offset of. Shape of (Y, X)
        ref_frame: reference frame to align frame to. Shape of (Y, X)
        
    Return:
        (dx, dy): best shift to shift frame to match reference frame
    """
    if guess_center is None:
        guess_center = np.array([frame.shape[1]/2., frame.shape[0]/2.])
    
    if outer_mask is None:
        outer_mask = frame.shape[0]/2 - 10
    
    # make mask
    y, x = np.indices(frame.shape)
    r = np.sqrt((x-guess_center[0])**2 + (y-guess_center[1])**2)
    mask = np.ones(frame.shape)
    bad = np.where((r < inner_mask) | (r > outer_mask))
    mask[bad] = 0
    
    result = scipy.optimize.minimize(_img_xcorr_shift, (0,0), args=(ref_frame, frame, mask), method="Nelder-Mead")
    
    return result.x




import abc
import os
import warnings
import subprocess
import multiprocessing as mp
import numpy as np
import astropy.io.fits as fits
import pyklip.klip as klip
import pyklip.instruments.utils.wcsgen as wcsgen

class Data(object):
    """
    Abstract Class with the required fields and methods that need to be implemented

    Attributes:
        input: Array of shape (N,y,x) for N images of shape (y,x)
        centers: Array of shape (N,2) for N input centers in the format [x_cent, y_cent]
        filenums: Array of size N for the numerical index to map data to file that was passed in
        filenames: Array of size N for the actual filepath of the file that corresponds to the data
        PAs: Array of N for the parallactic angle rotation of the target (used for ADI) [in degrees]
        wvs: Array of N wavelengths of the images (used for SDI) [in microns]. For polarization data, defaults to "None"
        wcs: Array of N wcs astormetry headers for each input image.
        IWA: a floating point scalar (not array). Specifies to inner working angle in pixels
        OWA: (optional) specifies outer working angle in pixels
        output: Array of shape (b, len(files), len(uniq_wvs), y, x) where b is the number of different KL basis cutoffs
        output_centers: Array of shape (N,2) for N output centers. Also coresponds to FM centers (does not need to be implemented)
        output_wcs: Array of N wcs astrometry headers for each output image (does not need to be implemneted)
        creator: (optional) string for creator of the data (used to identify pipelines that call pyklip)
        klipparams: (optional) a string that saves the most recent KLIP parameters
        flipx: (optional) False by default. Determines whether a relfection about the x axis is necessary to rotate image North-up East left

    Methods:
        readdata(): reread in the dadta
        savedata(): save a specified data in the GPI datacube format (in the 1st extension header)
        calibrate_output(): flux calibrate the output data
    """
    __metaclass__ = abc.ABCMeta

    def __init__(self):
        # set field for the creator of the data (used for pipeline work)
        self.creator = None
        # set field for klip parameters
        self.klipparams = None
        # set the outer working angle (optional parameter)
        self.OWA = None
        # determine whether a reflection is needed for North-up East-left (optional)
        self.flipx = False
        # self output centers and wcs to None until after running KLIP
        self.output_centers = None
        self.output_wcs = None


    ###################################
    ### Required Instance Variances ###
    ###################################

    #Note that each field has a getter and setter method so by default they are all read/write

    @abc.abstractproperty
    def input(self):
        """
        Input Data. Shape of (N, y, x)
        """
        return
    @input.setter
    def input(self, newval):
        return

    @abc.abstractproperty
    def centers(self):
        """
        Image centers. Shape of (N, 2) where the 2nd dimension is [x,y] pixel coordinate (in that order)
        """
        return
    @centers.setter
    def centers(self, newval):
        return

    @abc.abstractproperty
    def filenums(self):
        """
        Array of size N for the numerical index to map data to file that was passed in
        """
        return
    @filenums.setter
    def filenums(self, newval):
        return

    @abc.abstractproperty
    def filenames(self):
        """
        Array of size N for the actual filepath of the file that corresponds to the data
        """
        return
    @filenames.setter
    def filenames(self, newval):
        return


    @abc.abstractproperty
    def PAs(self):
        """
        Array of N for the parallactic angle rotation of the target (used for ADI) [in degrees]
        """
        return
    @PAs.setter
    def PAs(self, newval):
        return


    @abc.abstractproperty
    def wvs(self):
        """
        Array of N wavelengths (used for SDI) [in microns]. For polarization data, defaults to "None"
        """
        return
    @wvs.setter
    def wvs(self, newval):
        return


    @abc.abstractproperty
    def wcs(self):
        """
        Array of N wcs astormetry headers for each image.
        """
        return
    @wcs.setter
    def wcs(self, newval):
        return


    @abc.abstractproperty
    def IWA(self):
        """
        a floating point scalar (not array). Specifies to inner working angle in pixels
        """
        return
    @IWA.setter
    def IWA(self, newval):
        return


    @abc.abstractproperty
    def output(self):
        """
        Array of shape (b, len(files), len(uniq_wvs), y, x) where b is the number of different KL basis cutoffs
        """
        return
    @output.setter
    def output(self, newval):
        return


    # not an abstract property
    @property
    def numwvs(self):
        if not hasattr(self, "_numwvs"):
            self._numwvs = int(np.size(np.unique(self.wvs)))
        return self._numwvs


    ########################
    ### Required Methods ###
    ########################
    @abc.abstractmethod
    def readdata(self, filepaths):
        """
        Reads in the data from the files in the filelist and writes them to fields
        """
        return NotImplementedError("Subclass needs to implement this!")

    @staticmethod
    @abc.abstractmethod
    def savedata(self, filepath, data, klipparams=None, filetype="", zaxis=None, more_keywords=None):
        """
        Saves data for this instrument

        Args:
            filepath: filepath to save to
            data: data to save
            klipparams: a string of KLIP parameters. Write it to the 'PSFPARAM' keyword
            filtype: type of file (e.g. "KL Mode Cube", "PSF Subtracted Spectral Cube"). Wrriten to 'FILETYPE' keyword
            zaxis: a list of values for the zaxis of the datacub (for KL mode cubes currently)
            more_keywords (dictionary) : a dictionary {key: value, key:value} of header keywords and values which will
                                         written into the primary header
        """
        return NotImplementedError("Subclass needs to implement this!")

    @abc.abstractmethod
    def calibrate_output(self, img, spectral=False):
        """
        Calibrates the flux of an output image. Can either be a broadband image or a spectral cube depending
        on if the spectral flag is set.

        Assumes the broadband flux calibration is just multiplication by a single scalar number whereas spectral
        datacubes may have a separate calibration value for each wavelength

        Args:
            img: unclaibrated image.
                 If spectral is not set, this can either be a 2-D or 3-D broadband image
                 where the last two dimensions are [y,x]
                 If specetral is True, this is a 3-D spectral cube with shape [wv,y,x]
            spectral: if True, this is a spectral datacube. Otherwise, it is a broadband image.

        Return:
            calib_img: calibrated image of the same shape
        """
        return NotImplementedError("Subclass needs to implement this!")
    
    def spectral_collapse(self, collapse_channels=1, align_frames=True, aligned_center=None, numthreads=None, additional_params=None):
            """
            Collapses the dataset spectrally, bining the data into the desired number of output wavelengths. 
            This bins each cube individually; it does not bin the data tempoarally. 
            If number of wavelengths / output channels is not a whole number, some output channels will have more frames
            that went into the collapse

            Args:
                collapse_channels (int): number of output channels to evenly-ish collapse the dataset into. Default is 1 (broadband)
                align_frames (bool): if True, aligns each channel before collapse so that they are centered properly
                aligned_center: Array of shape (2) [x_cent, y_cent] for the centering the images to a given value
                numthreads (bool,int): number of threads to parallelize align and scale. If None, use default which is all of them
                additional_params (list of str): other dataset parameters to collapse. Assume each variable has first dimension of Nframes
            """
            # reshpae input into 4D cube
            Ncubes = self.input.shape[0] // self.numwvs
            input_4d = self.input.reshape([Ncubes, self.numwvs, self.input.shape[1], self.input.shape[2]])

            slices_per_group = self.numwvs // collapse_channels # how many wavelengths per each output channel
            leftover_slices = self.numwvs % collapse_channels

            collapsed_4d = np.zeros([Ncubes, collapse_channels, self.input.shape[1], self.input.shape[2]])
            wvs_collapsed = np.zeros([Ncubes, collapse_channels])
            pas_collapsed = np.zeros([Ncubes, collapse_channels])
            centers_collapsed = np.zeros([Ncubes, collapse_channels, 2])
            # appending following as lists
            wcs_collapsed = [] 
            filenums_collapsed = []
            filenames_collapsed = []
            # additional params, if needed
            if additional_params is not None:
                additional_collapsed = []
                for param_field in additional_params:
                    param_orig = getattr(self, param_field)
                    reshaped_shape = (Ncubes, collapse_channels) + param_orig.shape[1:]
                    additional_collapsed.append(np.zeros(reshaped_shape))
            # populate the output image
            next_start_channel = 0 # initialize which channel to start with for the input images
            for i in range(collapse_channels):
                # figure out which slices to pick
                slices_this_group = slices_per_group
                if leftover_slices > 0:
                    # take one extra slice, yummy
                    slices_this_group += 1
                    leftover_slices -= 1

                i_start = next_start_channel
                i_end = next_start_channel + slices_this_group # this is the index after the last one in this group

                if align_frames:
                    tpool = mp.Pool(processes=numthreads)

                    # for this range of wvs, one (x,y) center per cube
                    centers_4d = self.centers.reshape([Ncubes, self.numwvs, 2])
                    mean_centers =  np.mean(centers_4d[:,i_start:i_end,:], axis=1)
                    if aligned_center is not None:
                        mean_centers = mean_centers*0. + aligned_center

                    tasks = [tpool.apply_async(klip.align_and_scale, args=(img, new_center, old_center))        
                            for cube_j, new_center in enumerate(mean_centers)
                            for img, old_center in zip(input_4d[cube_j, i_start:i_end], centers_4d[cube_j, i_start:i_end])
                            ]

                    # reform back into a giant array
                    derotated = np.array([task.get() for task in tasks])
                    derotated.shape = (Ncubes, slices_this_group, self.input.shape[1], self.input.shape[2])
                    input_4d[:, i_start:i_end, :, :] = derotated

                # Remove annoying RuntimeWarnings when input_4d is all nans
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore", category=RuntimeWarning)
                    collapsed_4d[:,i,:,:] = np.nanmean(input_4d[:,i_start:i_end,:,:], axis=1)

                wvs_collapsed[:, i] = np.mean(self.wvs.reshape([Ncubes, self.numwvs])[:,i_start:i_end], axis=1)
                pas_collapsed[:, i] = np.mean(self.PAs.reshape([Ncubes, self.numwvs])[:,i_start:i_end], axis=1)
                centers_collapsed[:,i,:] = np.mean(self.centers.reshape([Ncubes, self.numwvs, 2])[:,i_start:i_end,:], axis=1)
                if aligned_center is not None:
                    centers_collapsed[:,i,:] = centers_collapsed[:,i,:]*0 + aligned_center
                
                # append arrays, we'll reshape them later
                # these variables are all the same for a single cube, so we can just select one
                wcs_collapsed.append(self.wcs.reshape([Ncubes, self.numwvs])[:,i_start]) 
                filenums_collapsed.append(self.filenums.reshape([Ncubes, self.numwvs])[:,i_start]) 
                
                if hasattr(self, 'irdis_rdp'):
                    filenames_collapsed = self.filenames
                else:
                    filenames_collapsed.append(self.filenames.reshape([Ncubes, self.numwvs])[:,i_start])

                if additional_params is not None:
                    for param_collapsed, param_field in zip(additional_collapsed, additional_params):
                        param_orig = getattr(self, param_field)
                        reshaped_shape = (Ncubes, self.numwvs) + param_orig.shape[1:]
                        param_collapsed[:, i] = np.nanmean(param_orig.reshape(reshaped_shape)[:, i_start:i_end], axis=1)

                next_start_channel = i_end

            # unravel the wavelength information
            collapsed_4d.shape = [Ncubes * collapse_channels, self.input.shape[1], self.input.shape[2]]
            wvs_collapsed.shape = [Ncubes * collapse_channels]
            pas_collapsed.shape = [Ncubes * collapse_channels]
            centers_collapsed.shape = [Ncubes * collapse_channels, 2]

            # unfold the lists, need to flip the dimensions, so they are ordered properly
            wcs_collapsed = np.array(wcs_collapsed).T.ravel()
            filenums_collapsed = np.array(filenums_collapsed).T.ravel()
            filenames_collapsed = np.array(filenames_collapsed).T.ravel()

            # ok time to set all the variables correctly
            self._numwvs = collapse_channels
            self.input = collapsed_4d
            self.wvs = wvs_collapsed
            self.PAs = pas_collapsed
            self.centers = centers_collapsed
            self.wcs = wcs_collapsed
            self.filenums = filenums_collapsed
            self.filenames = filenames_collapsed

            if additional_params is not None:
                for param_field, param_collapsed in zip(additional_params, additional_collapsed):
                    param_collapsed.shape = (Ncubes * collapse_channels, ) + param_collapsed.shape[2:]
                    setattr(self, param_field, param_collapsed)


class GenericData(Data):
    """
    Basic class to interface with a basic direct imaging dataset

    Args:
        input_data: either a 1-D list of filenames to read in, or a 3-D cube of all data (N, y, x)
        centers: array of shape (N,2) for N centers in the format [x_cent, y_cent]
        parangs: Array of N for the parallactic angle rotation of the target (used for ADI) [in degrees]
        wvs: Array of N wavelengths of the images (used for SDI) [in microns]. For polarization data, defaults to "None"
        IWA: a floating point scalar (not array). Specifies to inner working angle in pixels
        filenames: Array of size N for the actual filepath of the file that corresponds to the data
        flipx (boo): if True, the input images are right-handed (East clockwise of North) and need to be flipped for North-up-East-left

    Attributes:
        input: Array of shape (N,y,x) for N images of shape (y,x)
        centers: Array of shape (N,2) for N centers in the format [x_cent, y_cent]
        filenums: Array of size N for the numerical index to map data to file that was passed in
        filenames: Array of size N for the actual filepath of the file that corresponds to the data
        PAs: Array of N for the parallactic angle rotation of the target (used for ADI) [in degrees]
        wvs: Array of N wavelengths of the images (used for SDI) [in microns]. For polarization data, defaults to "None"
        wcs: Array of N wcs astormetry headers for each image.
        IWA: a floating point scalar (not array). Specifies to inner working angle in pixels
        output: Array of shape (b, len(files), len(uniq_wvs), y, x) where b is the number of different KL basis cutoffs
    """
    # Constructor
    def __init__(self, input_data, centers, parangs=None, wvs=None, IWA=0, filenames=None, flipx=False):
        super(GenericData, self).__init__()
        # read in the data
        if np.array(input_data).ndim == 1:
            self._input = self.readdata(input_data)
        else:
            # assume this is a 3-D cube
            self._input = np.array(input_data)
        
        nfiles = self.input.shape[0]

        self.centers = np.array(centers)

        if self.centers.shape [0] != nfiles:
            raise ValueError("Input data has shape {0} but centers has shape {1}".format(self.input.shape,
                                                                                         self.centers.shape))

        if parangs is not None:
            self._PAs = parangs
        else:
            self._PAs = np.zeros(nfiles)

        if wvs is not None:
            self._wvs = wvs
        else:
            self._wvs = np.ones(nfiles)

        self.IWA = IWA

        if filenames is not None:
            self._filenames = filenames
            unique_filenames = np.unique(filenames)                                                                                 
            self._filenums = np.array([np.argwhere(filename == unique_filenames).ravel()[0] for filename in filenames])
        else:
            self._filenums = np.arange(nfiles)
            self._filenames = np.array(["{0}".format(i) for i in self.filenums])

        self.flipx = flipx

        self._wcs = np.array([wcsgen.generate_wcs(parang, center, flipx=flipx) for parang, center in zip(self._PAs, self.centers)])

        self._output = None


         
    ################################
    ### Instance Required Fields ###
    ################################

    @property
    def input(self):
        return self._input
    @input.setter
    def input(self, newval):
        self._input = newval

    @property
    def centers(self):
        return self._centers
    @centers.setter
    def centers(self, newval):
        self._centers = newval

    @property
    def filenums(self):
        return self._filenums
    @filenums.setter
    def filenums(self, newval):
        self._filenums = newval

    @property
    def filenames(self):
        return self._filenames
    @filenames.setter
    def filenames(self, newval):
        self._filenames = newval

    @property
    def PAs(self):
        return self._PAs
    @PAs.setter
    def PAs(self, newval):
        self._PAs = newval

    @property
    def wvs(self):
        return self._wvs
    @wvs.setter
    def wvs(self, newval):
        self._wvs = newval

    @property
    def wcs(self):
        return self._wcs
    @wcs.setter
    def wcs(self, newval):
        self._wcs = newval


    @property
    def IWA(self):
        return self._IWA
    @IWA.setter
    def IWA(self, newval):
        self._IWA = newval

    @property
    def output(self):
        return self._output
    @output.setter
    def output(self, newval):
        self._output = newval
 

    def readdata(self, filepaths):
        """
        Reads in the data from the files in the filelist and writes them to fields.
        """
        input_data = []
        for filename in filepaths:
            with fits.open(filename) as hdulist:
                # assume the data is in the primary header
                data = hdulist[0].data
                # if this data has more than 2-D, collapse the Data
                dims = data.shape
                if np.size(dims) > 2:
                    nframes = np.prod(dims[:-2])
                    # collapse in all dimensions except y and x
                    data.shape = (nframes, dims[-2], dims[-1])

                input_data.append(data)

        # collapse data again
        input_data = np.array(input_data)
        dims = input_data.shape
        if np.size(dims) > 3:
            nframes = np.prod(dims[:-2])
            # collapse in all dimensions except y and x
            input_data.shape = (nframes, dims[-2], dims[-1])


    def savedata(self, filepath, data, klipparams=None, filetype="", zaxis=None, more_keywords=None):
        """
        Saves data for this instrument

        Args:
            filepath: filepath to save to
            data: data to save
            klipparams: a string of KLIP parameters. Write it to the 'PSFPARAM' keyword
            filtype: type of file (e.g. "KL Mode Cube", "PSF Subtracted Spectral Cube"). Wrriten to 'FILETYPE' keyword
            zaxis: a list of values for the zaxis of the datacub (for KL mode cubes currently)
            more_keywords (dictionary) : a dictionary {key: value, key:value} of header keywords and values which will
                                         written into the primary header
        """
        hdulist = fits.HDUList()
        hdulist.append(fits.PrimaryHDU(data=data))

        # save all the files we used in the reduction
        # we'll assume you used all the input files
        # remove duplicates from list
        filenames = np.unique(self.filenames)
        nfiles = np.size(filenames)
        hdulist[0].header["DRPNFILE"] = (nfiles, "Num raw files used in pyKLIP")
        for i, filename in enumerate(filenames):
            hdulist[0].header["FILE_{0}".format(i)] = filename + '.fits'

        # write out psf subtraction parameters
        # get pyKLIP revision number
        pykliproot = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
        # the universal_newline argument is just so python3 returns a string instead of bytes
        # this will probably come to bite me later
        try:
            pyklipver = subprocess.check_output(['git', 'rev-parse', '--short', 'HEAD'], cwd=pykliproot, universal_newlines=True).strip()
        except:
            pyklipver = "unknown"
        hdulist[0].header['PSFSUB'] = ("pyKLIP", "PSF Subtraction Algo")
        hdulist[0].header.add_history("Reduced with pyKLIP using commit {0}".format(pyklipver))
        hdulist[0].header['CREATOR'] = "pyKLIP-{0}".format(pyklipver)

        # store commit number for pyklip
        hdulist[0].header['pyklipv'] = (pyklipver, "pyKLIP version that was used")

        if klipparams is not None:
            hdulist[0].header['PSFPARAM'] = (klipparams, "KLIP parameters")
            hdulist[0].header.add_history("pyKLIP reduction with parameters {0}".format(klipparams))

        # write z axis units if necessary
        if zaxis is not None:
            # Writing a KL mode Cube
            if "KL Mode" in filetype:
                hdulist[0].header['CTYPE3'] = 'KLMODES'
                # write them individually
                for i, klmode in enumerate(zaxis):
                    hdulist[0].header['KLMODE{0}'.format(i)] = (klmode, "KL Mode of slice {0}".format(i))
                hdulist[0].header['CUNIT3'] = "N/A"
                hdulist[0].header['CRVAL3'] = 1
                hdulist[0].header['CRPIX3'] = 1.
                hdulist[0].header['CD3_3'] = 1.

        if "Spectral" in filetype:
            uniquewvs = np.unique(self.wvs)
            # do spectral stuff instead
            # because wavelength solutoin is nonlinear, we're not going to store it here
            hdulist[0].header['CTYPE3'] = 'WAVE'
            hdulist[0].header['CUNIT3'] = "N/A"
            hdulist[0].header['CRPIX3'] = 1.
            hdulist[0].header['CRVAL3'] = 0
            hdulist[0].header['CD3_3'] = 1
            # write it out instead
            for i, wv in enumerate(uniquewvs):
                hdulist[0].header['WV{0}'.format(i)] = (wv, "Wavelength of slice {0}".format(i))


        # store WCS information
        wcshdr = self.output_wcs[0].to_header()
        for key in wcshdr.keys():
            hdulist[0].header[key] = wcshdr[key]

        # but update the image center
        center = self.output_centers[0]
        hdulist[0].header.update({'PSFCENTX': center[0], 'PSFCENTY': center[1]})
        hdulist[0].header.update({'CRPIX1': center[0], 'CRPIX2': center[1]})
        hdulist[0].header.add_history("Image recentered to {0}".format(str(center)))

        if more_keywords is not None:
            hdulist[0].header.update(more_keywords)

        try:
            hdulist.writeto(filepath, overwrite=True)
        except TypeError:
            hdulist.writeto(filepath, clobber=True)
        hdulist.close()

    def calibrate_output(self, img, spectral=False):
        """
        Calibrates the flux of an output image. Can either be a broadband image or a spectral cube depending
        on if the spectral flag is set.

        Assumes the broadband flux calibration is just multiplication by a single scalar number whereas spectral
        datacubes may have a separate calibration value for each wavelength

        Args:
            img: unclaibrated image.
                 If spectral is not set, this can either be a 2-D or 3-D broadband image
                 where the last two dimensions are [y,x]
                 If specetral is True, this is a 3-D spectral cube with shape [wv,y,x]
            spectral: if True, this is a spectral datacube. Otherwise, it is a broadband image.

        Return:
            calib_img: calibrated image of the same shape
        """
        return img